import inspect
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Union

from jina import Document, DocumentArray
from jina._docarray import docarray_v2
from jina.importer import ImportExtensions
from jina.serve.networking.sse import EventSourceResponse
from jina.types.request.data import DataRequest


if TYPE_CHECKING:
    from jina.logging.logger import JinaLogger

if docarray_v2:
    from docarray import BaseDoc, DocList


def get_fastapi_app(
    request_models_map: Dict,
    caller: Callable,
    logger: 'JinaLogger',
    cors: bool = False,
    **kwargs,
):
    """
    Get the app from FastAPI as the REST interface.

    :param request_models_map: Map describing the endpoints and its Pydantic models
    :param caller: Callable to be handled by the endpoints of the returned FastAPI app
    :param logger: Logger object
    :param cors: If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access.
    :param kwargs: Extra kwargs to make it compatible with other methods
    :return: fastapi app
    """
    with ImportExtensions(required=True):
        from fastapi import FastAPI, Response, HTTPException
        import pydantic
        from fastapi.middleware.cors import CORSMiddleware
    import os

    from pydantic import BaseModel, Field
    from pydantic.config import BaseConfig, inherit_config

    from jina.proto import jina_pb2
    from jina.serve.runtimes.gateway.models import _to_camel_case

    class Header(BaseModel):
        request_id: Optional[str] = Field(
            description='Request ID', example=os.urandom(16).hex()
        )

        class Config(BaseConfig):
            alias_generator = _to_camel_case
            allow_population_by_field_name = True

    class InnerConfig(BaseConfig):
        alias_generator = _to_camel_case
        allow_population_by_field_name = True

    app = FastAPI()

    if cors:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*'],
        )
        logger.warning('CORS is enabled. This service is accessible from any website!')

    def add_post_route(
        endpoint_path,
        input_model,
        output_model,
        input_doc_list_model=None,
        output_doc_list_model=None,
    ):
        app_kwargs = dict(
            path=f'/{endpoint_path.strip("/")}',
            methods=['POST'],
            summary=f'Endpoint {endpoint_path}',
            response_model=output_model,
        )
        if docarray_v2:
            from docarray.base_doc.docarray_response import DocArrayResponse

            app_kwargs['response_class'] = DocArrayResponse

        @app.api_route(**app_kwargs)
        async def post(body: input_model, response: Response):
            req = DataRequest()
            if body.header is not None:
                req.header.request_id = body.header.request_id

            if body.parameters is not None:
                req.parameters = body.parameters
            req.header.exec_endpoint = endpoint_path
            data = body.data
            if isinstance(data, list):
                if not docarray_v2:
                    req.direct_docs = DocumentArray.from_pydantic_model(data)
                else:
                    req.document_array_cls = DocList[input_doc_model]
                    req.direct_docs = DocList[input_doc_list_model](data)
            else:
                if not docarray_v2:
                    req.direct_docs = DocumentArray([Document.from_pydantic_model(data)])
                else:
                    req.document_array_cls = DocList[input_doc_model]
                    req.direct_docs = DocList[input_doc_list_model]([data])
                if body.header is None:
                    req.header.request_id = req.docs[0].id

            resp = await caller(req)
            status = resp.header.status

            if status.code == jina_pb2.StatusProto.ERROR:
                raise HTTPException(status_code=499, detail=status.description)
            else:
                if not docarray_v2:
                    docs_response = resp.docs.to_dict()
                else:
                    docs_response = resp.docs
                ret = output_model(data=docs_response, parameters=resp.parameters)

                return ret

    def add_streaming_routes(
        endpoint_path,
        input_doc_model=None,
    ):
        from fastapi import Request

        @app.api_route(
            path=f'/{endpoint_path.strip("/")}',
            methods=['GET'],
            summary=f'Streaming Endpoint {endpoint_path}',
        )
        async def streaming_get(request: Request = None, body: input_doc_model = None):
            if not body:
                query_params = dict(request.query_params)
                body = (
                    input_doc_model.parse_obj(query_params)
                    if docarray_v2
                    else Document.from_dict(query_params)
                )
            else:
                if not docarray_v2:
                    body = Document.from_pydantic_model(body)
            req = DataRequest()
            req.header.exec_endpoint = endpoint_path
            if not docarray_v2:
                req.direct_docs = DocumentArray([body])
            else:
                req.document_array_cls = DocList[input_doc_model]
                req.direct_docs = DocList[input_doc_model]([body])
            event_generator = _gen_dict_documents(await caller(req))
            return EventSourceResponse(event_generator)

    for endpoint, input_output_map in request_models_map.items():
        if endpoint != '_jina_dry_run_':
            input_doc_model = input_output_map['input']['model']
            output_doc_model = input_output_map['output']['model']
            is_generator = input_output_map['is_generator']
            parameters_model = input_output_map['parameters']['model']
            parameters_model_needed = parameters_model is not None
            if parameters_model_needed:
                try:
                    _ = parameters_model()
                    parameters_model_needed = False
                except:
                    parameters_model_needed = True
                parameters_model = parameters_model if parameters_model_needed else Optional[parameters_model]
                default_parameters = (
                    ... if parameters_model_needed else None
                )
            else:
                parameters_model = Optional[Dict]
                default_parameters = None

            if docarray_v2:
                _config = inherit_config(InnerConfig, BaseDoc.__config__)
            else:
                _config = input_doc_model.__config__

            endpoint_input_model = pydantic.create_model(
                f'{endpoint.strip("/")}_input_model',
                data=(Union[List[input_doc_model], input_doc_model], ...),
                parameters=(parameters_model, default_parameters),
                header=(Optional[Header], None),
                __config__=_config,
            )

            endpoint_output_model = pydantic.create_model(
                f'{endpoint.strip("/")}_output_model',
                data=(Union[List[output_doc_model], output_doc_model], ...),
                parameters=(Optional[Dict], None),
                __config__=_config,
            )

            if is_generator:
                add_streaming_routes(
                    endpoint,
                    input_doc_model=input_doc_model,
                )
            else:
                add_post_route(
                    endpoint,
                    input_model=endpoint_input_model,
                    output_model=endpoint_output_model,
                    input_doc_list_model=input_doc_model,
                    output_doc_list_model=output_doc_model,
                )

    from jina.serve.runtimes.gateway.health_model import JinaHealthModel

    @app.get(
        path='/',
        summary='Get the health of Jina Executor service',
        response_model=JinaHealthModel,
    )
    async def _executor_health():
        """
        Get the health of this Gateway service.
        .. # noqa: DAR201

        """
        return {}

    return app


async def _gen_dict_documents(gen):
    if inspect.isasyncgen(gen):
        async for document in gen:
            yield _doc_to_event(document)
    else:
        for document in gen:
            yield _doc_to_event(document)
    yield {'event': 'end'}


def _doc_to_event(doc):
    if not docarray_v2:
        return {'event': 'update', 'data': doc.to_dict()}
    else:
        return {'event': 'update', 'data': doc.dict()}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: jina.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import docarray.proto.pb2.docarray_pb2 as docarray__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\njina.proto\x12\x04jina\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0e\x64ocarray.proto\"\x9f\x01\n\nRouteProto\x12\x10\n\x08\x65xecutor\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12!\n\x06status\x18\x04 \x01(\x0b\x32\x11.jina.StatusProto\"\xc3\x01\n\rJinaInfoProto\x12+\n\x04jina\x18\x01 \x03(\x0b\x32\x1d.jina.JinaInfoProto.JinaEntry\x12+\n\x04\x65nvs\x18\x02 \x03(\x0b\x32\x1d.jina.JinaInfoProto.EnvsEntry\x1a+\n\tJinaEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a+\n\tEnvsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xc6\x01\n\x0bHeaderProto\x12\x12\n\nrequest_id\x18\x01 \x01(\t\x12!\n\x06status\x18\x02 \x01(\x0b\x32\x11.jina.StatusProto\x12\x1a\n\rexec_endpoint\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x1c\n\x0ftarget_executor\x18\x04 \x01(\tH\x01\x88\x01\x01\x12\x14\n\x07timeout\x18\x05 \x01(\rH\x02\x88\x01\x01\x42\x10\n\x0e_exec_endpointB\x12\n\x10_target_executorB\n\n\x08_timeout\"f\n\x0e\x45ndpointsProto\x12\x11\n\tendpoints\x18\x01 \x03(\t\x12\x17\n\x0fwrite_endpoints\x18\x02 \x03(\t\x12(\n\x07schemas\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\"\xf9\x01\n\x0bStatusProto\x12*\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x1c.jina.StatusProto.StatusCode\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x33\n\texception\x18\x03 \x01(\x0b\x32 .jina.StatusProto.ExceptionProto\x1aN\n\x0e\x45xceptionProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\t\x12\x0e\n\x06stacks\x18\x03 \x03(\t\x12\x10\n\x08\x65xecutor\x18\x04 \x01(\t\"$\n\nStatusCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05\x45RROR\x10\x01\"^\n\rRelatedEntity\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\r\x12\x15\n\x08shard_id\x18\x04 \x01(\rH\x00\x88\x01\x01\x42\x0b\n\t_shard_id\"\xa0\x02\n\x10\x44\x61taRequestProto\x12!\n\x06header\x18\x01 \x01(\x0b\x32\x11.jina.HeaderProto\x12+\n\nparameters\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12 \n\x06routes\x18\x03 \x03(\x0b\x32\x10.jina.RouteProto\x12\x35\n\x04\x64\x61ta\x18\x04 \x01(\x0b\x32\'.jina.DataRequestProto.DataContentProto\x1a\x63\n\x10\x44\x61taContentProto\x12,\n\x04\x64ocs\x18\x01 \x01(\x0b\x32\x1c.docarray.DocumentArrayProtoH\x00\x12\x14\n\ndocs_bytes\x18\x02 \x01(\x0cH\x00\x42\x0b\n\tdocuments\"\xb9\x01\n\x1aSingleDocumentRequestProto\x12!\n\x06header\x18\x01 \x01(\x0b\x32\x11.jina.HeaderProto\x12+\n\nparameters\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12 \n\x06routes\x18\x03 \x03(\x0b\x32\x10.jina.RouteProto\x12)\n\x08\x64ocument\x18\x04 \x01(\x0b\x32\x17.docarray.DocumentProto\"\x8a\x01\n\x16\x44\x61taRequestProtoWoData\x12!\n\x06header\x18\x01 \x01(\x0b\x32\x11.jina.HeaderProto\x12+\n\nparameters\x18\x02 \x01(\x0b\x32\x17.google.protobuf.Struct\x12 \n\x06routes\x18\x03 \x03(\x0b\x32\x10.jina.RouteProto\"@\n\x14\x44\x61taRequestListProto\x12(\n\x08requests\x18\x01 \x03(\x0b\x32\x16.jina.DataRequestProto\"\x1b\n\nSnapshotId\x12\r\n\x05value\x18\x01 \x01(\t\"\x1a\n\tRestoreId\x12\r\n\x05value\x18\x01 \x01(\t\"\xef\x01\n\x13SnapshotStatusProto\x12\x1c\n\x02id\x18\x01 \x01(\x0b\x32\x10.jina.SnapshotId\x12\x30\n\x06status\x18\x02 \x01(\x0e\x32 .jina.SnapshotStatusProto.Status\x12\x15\n\rsnapshot_file\x18\x03 \x01(\t\"q\n\x06Status\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\n\n\x06QUEUED\x10\x01\x12\r\n\tSCHEDULED\x10\x02\x12\x0b\n\x07RUNNING\x10\x03\x12\r\n\tSUCCEEDED\x10\x04\x12\n\n\x06\x46\x41ILED\x10\x05\x12\r\n\tNOT_FOUND\x10\x06\"\xca\x01\n\x1aRestoreSnapshotStatusProto\x12\x1b\n\x02id\x18\x01 \x01(\x0b\x32\x0f.jina.RestoreId\x12\x37\n\x06status\x18\x02 \x01(\x0e\x32\'.jina.RestoreSnapshotStatusProto.Status\"V\n\x06Status\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07RUNNING\x10\x01\x12\r\n\tSUCCEEDED\x10\x02\x12\n\n\x06\x46\x41ILED\x10\x03\x12\r\n\tNOT_FOUND\x10\x06\"/\n\x16RestoreSnapshotCommand\x12\x15\n\rsnapshot_file\x18\x01 \x01(\t2Z\n\x12JinaDataRequestRPC\x12\x44\n\x0cprocess_data\x12\x1a.jina.DataRequestListProto\x1a\x16.jina.DataRequestProto\"\x00\x32\x63\n\x18JinaSingleDataRequestRPC\x12G\n\x13process_single_data\x12\x16.jina.DataRequestProto\x1a\x16.jina.DataRequestProto\"\x00\x32t\n\x1cJinaSingleDocumentRequestRPC\x12T\n\nstream_doc\x12 .jina.SingleDocumentRequestProto\x1a .jina.SingleDocumentRequestProto\"\x00\x30\x01\x32G\n\x07JinaRPC\x12<\n\x04\x43\x61ll\x12\x16.jina.DataRequestProto\x1a\x16.jina.DataRequestProto\"\x00(\x01\x30\x01\x32`\n\x18JinaDiscoverEndpointsRPC\x12\x44\n\x12\x65ndpoint_discovery\x12\x16.google.protobuf.Empty\x1a\x14.jina.EndpointsProto\"\x00\x32N\n\x14JinaGatewayDryRunRPC\x12\x36\n\x07\x64ry_run\x12\x16.google.protobuf.Empty\x1a\x11.jina.StatusProto\"\x00\x32G\n\x0bJinaInfoRPC\x12\x38\n\x07_status\x12\x16.google.protobuf.Empty\x1a\x13.jina.JinaInfoProto\"\x00\x32W\n\x14JinaExecutorSnapshot\x12?\n\x08snapshot\x12\x16.google.protobuf.Empty\x1a\x19.jina.SnapshotStatusProto\"\x00\x32`\n\x1cJinaExecutorSnapshotProgress\x12@\n\x0fsnapshot_status\x12\x10.jina.SnapshotId\x1a\x19.jina.SnapshotStatusProto\"\x00\x32\x62\n\x13JinaExecutorRestore\x12K\n\x07restore\x12\x1c.jina.RestoreSnapshotCommand\x1a .jina.RestoreSnapshotStatusProto\"\x00\x32\x64\n\x1bJinaExecutorRestoreProgress\x12\x45\n\x0erestore_status\x12\x0f.jina.RestoreId\x1a .jina.RestoreSnapshotStatusProto\"\x00\x62\x06proto3'
)


_ROUTEPROTO = DESCRIPTOR.message_types_by_name['RouteProto']
_JINAINFOPROTO = DESCRIPTOR.message_types_by_name['JinaInfoProto']
_JINAINFOPROTO_JINAENTRY = _JINAINFOPROTO.nested_types_by_name['JinaEntry']
_JINAINFOPROTO_ENVSENTRY = _JINAINFOPROTO.nested_types_by_name['EnvsEntry']
_HEADERPROTO = DESCRIPTOR.message_types_by_name['HeaderProto']
_ENDPOINTSPROTO = DESCRIPTOR.message_types_by_name['EndpointsProto']
_STATUSPROTO = DESCRIPTOR.message_types_by_name['StatusProto']
_STATUSPROTO_EXCEPTIONPROTO = _STATUSPROTO.nested_types_by_name['ExceptionProto']
_RELATEDENTITY = DESCRIPTOR.message_types_by_name['RelatedEntity']
_DATAREQUESTPROTO = DESCRIPTOR.message_types_by_name['DataRequestProto']
_DATAREQUESTPROTO_DATACONTENTPROTO = _DATAREQUESTPROTO.nested_types_by_name[
    'DataContentProto'
]
_SINGLEDOCUMENTREQUESTPROTO = DESCRIPTOR.message_types_by_name[
    'SingleDocumentRequestProto'
]
_DATAREQUESTPROTOWODATA = DESCRIPTOR.message_types_by_name['DataRequestProtoWoData']
_DATAREQUESTLISTPROTO = DESCRIPTOR.message_types_by_name['DataRequestListProto']
_SNAPSHOTID = DESCRIPTOR.message_types_by_name['SnapshotId']
_RESTOREID = DESCRIPTOR.message_types_by_name['RestoreId']
_SNAPSHOTSTATUSPROTO = DESCRIPTOR.message_types_by_name['SnapshotStatusProto']
_RESTORESNAPSHOTSTATUSPROTO = DESCRIPTOR.message_types_by_name[
    'RestoreSnapshotStatusProto'
]
_RESTORESNAPSHOTCOMMAND = DESCRIPTOR.message_types_by_name['RestoreSnapshotCommand']
_STATUSPROTO_STATUSCODE = _STATUSPROTO.enum_types_by_name['StatusCode']
_SNAPSHOTSTATUSPROTO_STATUS = _SNAPSHOTSTATUSPROTO.enum_types_by_name['Status']
_RESTORESNAPSHOTSTATUSPROTO_STATUS = _RESTORESNAPSHOTSTATUSPROTO.enum_types_by_name[
    'Status'
]
RouteProto = _reflection.GeneratedProtocolMessageType(
    'RouteProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _ROUTEPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.RouteProto)
    },
)
_sym_db.RegisterMessage(RouteProto)

JinaInfoProto = _reflection.GeneratedProtocolMessageType(
    'JinaInfoProto',
    (_message.Message,),
    {
        'JinaEntry': _reflection.GeneratedProtocolMessageType(
            'JinaEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _JINAINFOPROTO_JINAENTRY,
                '__module__': 'jina_pb2',
                # @@protoc_insertion_point(class_scope:jina.JinaInfoProto.JinaEntry)
            },
        ),
        'EnvsEntry': _reflection.GeneratedProtocolMessageType(
            'EnvsEntry',
            (_message.Message,),
            {
                'DESCRIPTOR': _JINAINFOPROTO_ENVSENTRY,
                '__module__': 'jina_pb2',
                # @@protoc_insertion_point(class_scope:jina.JinaInfoProto.EnvsEntry)
            },
        ),
        'DESCRIPTOR': _JINAINFOPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.JinaInfoProto)
    },
)
_sym_db.RegisterMessage(JinaInfoProto)
_sym_db.RegisterMessage(JinaInfoProto.JinaEntry)
_sym_db.RegisterMessage(JinaInfoProto.EnvsEntry)

HeaderProto = _reflection.GeneratedProtocolMessageType(
    'HeaderProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _HEADERPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.HeaderProto)
    },
)
_sym_db.RegisterMessage(HeaderProto)

EndpointsProto = _reflection.GeneratedProtocolMessageType(
    'EndpointsProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _ENDPOINTSPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.EndpointsProto)
    },
)
_sym_db.RegisterMessage(EndpointsProto)

StatusProto = _reflection.GeneratedProtocolMessageType(
    'StatusProto',
    (_message.Message,),
    {
        'ExceptionProto': _reflection.GeneratedProtocolMessageType(
            'ExceptionProto',
            (_message.Message,),
            {
                'DESCRIPTOR': _STATUSPROTO_EXCEPTIONPROTO,
                '__module__': 'jina_pb2',
                # @@protoc_insertion_point(class_scope:jina.StatusProto.ExceptionProto)
            },
        ),
        'DESCRIPTOR': _STATUSPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.StatusProto)
    },
)
_sym_db.RegisterMessage(StatusProto)
_sym_db.RegisterMessage(StatusProto.ExceptionProto)

RelatedEntity = _reflection.GeneratedProtocolMessageType(
    'RelatedEntity',
    (_message.Message,),
    {
        'DESCRIPTOR': _RELATEDENTITY,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.RelatedEntity)
    },
)
_sym_db.RegisterMessage(RelatedEntity)

DataRequestProto = _reflection.GeneratedProtocolMessageType(
    'DataRequestProto',
    (_message.Message,),
    {
        'DataContentProto': _reflection.GeneratedProtocolMessageType(
            'DataContentProto',
            (_message.Message,),
            {
                'DESCRIPTOR': _DATAREQUESTPROTO_DATACONTENTPROTO,
                '__module__': 'jina_pb2',
                # @@protoc_insertion_point(class_scope:jina.DataRequestProto.DataContentProto)
            },
        ),
        'DESCRIPTOR': _DATAREQUESTPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.DataRequestProto)
    },
)
_sym_db.RegisterMessage(DataRequestProto)
_sym_db.RegisterMessage(DataRequestProto.DataContentProto)

SingleDocumentRequestProto = _reflection.GeneratedProtocolMessageType(
    'SingleDocumentRequestProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _SINGLEDOCUMENTREQUESTPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.SingleDocumentRequestProto)
    },
)
_sym_db.RegisterMessage(SingleDocumentRequestProto)

DataRequestProtoWoData = _reflection.GeneratedProtocolMessageType(
    'DataRequestProtoWoData',
    (_message.Message,),
    {
        'DESCRIPTOR': _DATAREQUESTPROTOWODATA,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.DataRequestProtoWoData)
    },
)
_sym_db.RegisterMessage(DataRequestProtoWoData)

DataRequestListProto = _reflection.GeneratedProtocolMessageType(
    'DataRequestListProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _DATAREQUESTLISTPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.DataRequestListProto)
    },
)
_sym_db.RegisterMessage(DataRequestListProto)

SnapshotId = _reflection.GeneratedProtocolMessageType(
    'SnapshotId',
    (_message.Message,),
    {
        'DESCRIPTOR': _SNAPSHOTID,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.SnapshotId)
    },
)
_sym_db.RegisterMessage(SnapshotId)

RestoreId = _reflection.GeneratedProtocolMessageType(
    'RestoreId',
    (_message.Message,),
    {
        'DESCRIPTOR': _RESTOREID,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.RestoreId)
    },
)
_sym_db.RegisterMessage(RestoreId)

SnapshotStatusProto = _reflection.GeneratedProtocolMessageType(
    'SnapshotStatusProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _SNAPSHOTSTATUSPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.SnapshotStatusProto)
    },
)
_sym_db.RegisterMessage(SnapshotStatusProto)

RestoreSnapshotStatusProto = _reflection.GeneratedProtocolMessageType(
    'RestoreSnapshotStatusProto',
    (_message.Message,),
    {
        'DESCRIPTOR': _RESTORESNAPSHOTSTATUSPROTO,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.RestoreSnapshotStatusProto)
    },
)
_sym_db.RegisterMessage(RestoreSnapshotStatusProto)

RestoreSnapshotCommand = _reflection.GeneratedProtocolMessageType(
    'RestoreSnapshotCommand',
    (_message.Message,),
    {
        'DESCRIPTOR': _RESTORESNAPSHOTCOMMAND,
        '__module__': 'jina_pb2',
        # @@protoc_insertion_point(class_scope:jina.RestoreSnapshotCommand)
    },
)
_sym_db.RegisterMessage(RestoreSnapshotCommand)

_JINADATAREQUESTRPC = DESCRIPTOR.services_by_name['JinaDataRequestRPC']
_JINASINGLEDATAREQUESTRPC = DESCRIPTOR.services_by_name['JinaSingleDataRequestRPC']
_JINASINGLEDOCUMENTREQUESTRPC = DESCRIPTOR.services_by_name[
    'JinaSingleDocumentRequestRPC'
]
_JINARPC = DESCRIPTOR.services_by_name['JinaRPC']
_JINADISCOVERENDPOINTSRPC = DESCRIPTOR.services_by_name['JinaDiscoverEndpointsRPC']
_JINAGATEWAYDRYRUNRPC = DESCRIPTOR.services_by_name['JinaGatewayDryRunRPC']
_JINAINFORPC = DESCRIPTOR.services_by_name['JinaInfoRPC']
_JINAEXECUTORSNAPSHOT = DESCRIPTOR.services_by_name['JinaExecutorSnapshot']
_JINAEXECUTORSNAPSHOTPROGRESS = DESCRIPTOR.services_by_name[
    'JinaExecutorSnapshotProgress'
]
_JINAEXECUTORRESTORE = DESCRIPTOR.services_by_name['JinaExecutorRestore']
_JINAEXECUTORRESTOREPROGRESS = DESCRIPTOR.services_by_name[
    'JinaExecutorRestoreProgress'
]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _JINAINFOPROTO_JINAENTRY._options = None
    _JINAINFOPROTO_JINAENTRY._serialized_options = b'8\001'
    _JINAINFOPROTO_ENVSENTRY._options = None
    _JINAINFOPROTO_ENVSENTRY._serialized_options = b'8\001'
    _ROUTEPROTO._serialized_start = 129
    _ROUTEPROTO._serialized_end = 288
    _JINAINFOPROTO._serialized_start = 291
    _JINAINFOPROTO._serialized_end = 486
    _JINAINFOPROTO_JINAENTRY._serialized_start = 398
    _JINAINFOPROTO_JINAENTRY._serialized_end = 441
    _JINAINFOPROTO_ENVSENTRY._serialized_start = 443
    _JINAINFOPROTO_ENVSENTRY._serialized_end = 486
    _HEADERPROTO._serialized_start = 489
    _HEADERPROTO._serialized_end = 687
    _ENDPOINTSPROTO._serialized_start = 689
    _ENDPOINTSPROTO._serialized_end = 791
    _STATUSPROTO._serialized_start = 794
    _STATUSPROTO._serialized_end = 1043
    _STATUSPROTO_EXCEPTIONPROTO._serialized_start = 927
    _STATUSPROTO_EXCEPTIONPROTO._serialized_end = 1005
    _STATUSPROTO_STATUSCODE._serialized_start = 1007
    _STATUSPROTO_STATUSCODE._serialized_end = 1043
    _RELATEDENTITY._serialized_start = 1045
    _RELATEDENTITY._serialized_end = 1139
    _DATAREQUESTPROTO._serialized_start = 1142
    _DATAREQUESTPROTO._serialized_end = 1430
    _DATAREQUESTPROTO_DATACONTENTPROTO._serialized_start = 1331
    _DATAREQUESTPROTO_DATACONTENTPROTO._serialized_end = 1430
    _SINGLEDOCUMENTREQUESTPROTO._serialized_start = 1433
    _SINGLEDOCUMENTREQUESTPROTO._serialized_end = 1618
    _DATAREQUESTPROTOWODATA._serialized_start = 1621
    _DATAREQUESTPROTOWODATA._serialized_end = 1759
    _DATAREQUESTLISTPROTO._serialized_start = 1761
    _DATAREQUESTLISTPROTO._serialized_end = 1825
    _SNAPSHOTID._serialized_start = 1827
    _SNAPSHOTID._serialized_end = 1854
    _RESTOREID._serialized_start = 1856
    _RESTOREID._serialized_end = 1882
    _SNAPSHOTSTATUSPROTO._serialized_start = 1885
    _SNAPSHOTSTATUSPROTO._serialized_end = 2124
    _SNAPSHOTSTATUSPROTO_STATUS._serialized_start = 2011
    _SNAPSHOTSTATUSPROTO_STATUS._serialized_end = 2124
    _RESTORESNAPSHOTSTATUSPROTO._serialized_start = 2127
    _RESTORESNAPSHOTSTATUSPROTO._serialized_end = 2329
    _RESTORESNAPSHOTSTATUSPROTO_STATUS._serialized_start = 2243
    _RESTORESNAPSHOTSTATUSPROTO_STATUS._serialized_end = 2329
    _RESTORESNAPSHOTCOMMAND._serialized_start = 2331
    _RESTORESNAPSHOTCOMMAND._serialized_end = 2378
    _JINADATAREQUESTRPC._serialized_start = 2380
    _JINADATAREQUESTRPC._serialized_end = 2470
    _JINASINGLEDATAREQUESTRPC._serialized_start = 2472
    _JINASINGLEDATAREQUESTRPC._serialized_end = 2571
    _JINASINGLEDOCUMENTREQUESTRPC._serialized_start = 2573
    _JINASINGLEDOCUMENTREQUESTRPC._serialized_end = 2689
    _JINARPC._serialized_start = 2691
    _JINARPC._serialized_end = 2762
    _JINADISCOVERENDPOINTSRPC._serialized_start = 2764
    _JINADISCOVERENDPOINTSRPC._serialized_end = 2860
    _JINAGATEWAYDRYRUNRPC._serialized_start = 2862
    _JINAGATEWAYDRYRUNRPC._serialized_end = 2940
    _JINAINFORPC._serialized_start = 2942
    _JINAINFORPC._serialized_end = 3013
    _JINAEXECUTORSNAPSHOT._serialized_start = 3015
    _JINAEXECUTORSNAPSHOT._serialized_end = 3102
    _JINAEXECUTORSNAPSHOTPROGRESS._serialized_start = 3104
    _JINAEXECUTORSNAPSHOTPROGRESS._serialized_end = 3200
    _JINAEXECUTORRESTORE._serialized_start = 3202
    _JINAEXECUTORRESTORE._serialized_end = 3300
    _JINAEXECUTORRESTOREPROGRESS._serialized_start = 3302
    _JINAEXECUTORRESTOREPROGRESS._serialized_end = 3402
# @@protoc_insertion_point(module_scope)



from jina.serve.runtimes.gateway.http.fastapi import (
    FastAPIBaseGateway,
)  # keep import here for backwards compatibility
from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.servers.http import HTTPServer

__all__ = ['HTTPGateway']


class HTTPGateway(HTTPServer, BaseGateway):
    """
    :class:`HTTPGateway` is a FastAPIBaseGateway that uses the default FastAPI app
    """

    pass



import logging
import os
from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

from jina._docarray import docarray_v2
from jina.importer import ImportExtensions
from jina.serve.runtimes.servers import BaseServer

if TYPE_CHECKING:
    from fastapi import FastAPI


class FastAPIBaseServer(BaseServer):
    """Base FastAPI server. Implement this abstract class in-case you want to build a fastapi-based server by
    implementing the `app` property. This property should return a fastapi app. The base Gateway will handle starting
    a server and serving the application using that server."""

    def __init__(
        self,
        ssl_keyfile: Optional[str] = None,
        ssl_certfile: Optional[str] = None,
        uvicorn_kwargs: Optional[dict] = None,
        proxy: bool = False,
        title: Optional[str] = None,
        description: Optional[str] = None,
        no_debug_endpoints: Optional[bool] = False,
        no_crud_endpoints: Optional[bool] = False,
        expose_endpoints: Optional[str] = None,
        expose_graphql_endpoint: Optional[bool] = False,
        cors: Optional[bool] = False,
        **kwargs,
    ):
        """Initialize the FastAPIBaseGateway
        :param ssl_keyfile: the path to the key file
        :param ssl_certfile: the path to the certificate file
        :param uvicorn_kwargs: Dictionary of kwargs arguments that will be passed to Uvicorn server when starting the server
        :param proxy: If set, respect the http_proxy and https_proxy environment variables, otherwise, it will unset
            these proxy variables before start. gRPC seems to prefer no proxy
        :param title: The title of this HTTP server. It will be used in automatics docs such as Swagger UI.
        :param description: The description of this HTTP server. It will be used in automatics docs such as Swagger UI.
        :param no_debug_endpoints: If set, `/status` `/post` endpoints are removed from HTTP interface.
        :param no_crud_endpoints: If set, `/index`, `/search`, `/update`, `/delete` endpoints are removed from HTTP interface.
                  Any executor that has `@requests(on=...)` bound with those values will receive data requests.
        :param expose_endpoints: A JSON string that represents a map from executor endpoints (`@requests(on=...)`) to HTTP endpoints.
        :param expose_graphql_endpoint: If set, /graphql endpoint is added to HTTP interface.
        :param cors: If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access.
        :param kwargs: keyword args
        """
        super().__init__(**kwargs)
        self.title = title
        self.description = description
        self.no_debug_endpoints = no_debug_endpoints
        self.no_crud_endpoints = no_crud_endpoints
        self.expose_endpoints = expose_endpoints
        self.expose_graphql_endpoint = expose_graphql_endpoint
        self.cors = cors
        self.uvicorn_kwargs = uvicorn_kwargs or {}

        if ssl_keyfile and 'ssl_keyfile' not in self.uvicorn_kwargs.keys():
            self.uvicorn_kwargs['ssl_keyfile'] = ssl_keyfile

        if ssl_certfile and 'ssl_certfile' not in self.uvicorn_kwargs.keys():
            self.uvicorn_kwargs['ssl_certfile'] = ssl_certfile

        if not proxy and os.name != 'nt':
            os.unsetenv('http_proxy')
            os.unsetenv('https_proxy')

    @property
    @abstractmethod
    def app(self):
        """Get a FastAPI app"""
        ...

    async def setup_server(self):
        """
        Initialize and return GRPC server
        """
        self.logger.debug(f'Setting up HTTP server')
        with ImportExtensions(required=True):
            from uvicorn import Config, Server

        class UviServer(Server):
            """The uvicorn server."""

            async def setup(self, sockets=None):
                """
                Setup uvicorn server.

                :param sockets: sockets of server.
                """
                config = self.config
                if not config.loaded:
                    config.load()
                self.lifespan = config.lifespan_class(config)
                await self.startup(sockets=sockets)
                if self.should_exit:
                    return

            async def serve(self, **kwargs):
                """
                Start the server.

                :param kwargs: keyword arguments
                """
                await self.main_loop()

        if 'CICD_JINA_DISABLE_HEALTHCHECK_LOGS' in os.environ:

            class _EndpointFilter(logging.Filter):
                def filter(self, record: logging.LogRecord) -> bool:
                    # NOTE: space is important after `GET /`, else all logs will be disabled.
                    return record.getMessage().find("GET / ") == -1

            # Filter out healthcheck endpoint `GET /`
            logging.getLogger("uvicorn.access").addFilter(_EndpointFilter())

        if docarray_v2:
            from jina.serve.runtimes.gateway.request_handling import (
                GatewayRequestHandler,
            )

            if isinstance(self._request_handler, GatewayRequestHandler):
                await self._request_handler.streamer._get_endpoints_input_output_models(
                    is_cancel=self.is_cancel
                )
                self._request_handler.streamer._validate_flow_docarray_compatibility()

        # app property will generate a new fastapi app each time called
        app = self.app
        _install_health_check(app, self.logger)
        self.server = UviServer(
            config=Config(
                app=app,
                host=self.host,
                port=self.port,
                log_level=os.getenv('JINA_LOG_LEVEL', 'error').lower(),
                **self.uvicorn_kwargs,
            )
        )
        self.logger.debug(f'UviServer server setup on port {self.port}')
        await self.server.setup()
        self.logger.debug(f'HTTP server setup successful')

    async def shutdown(self):
        """
        Free resources allocated when setting up HTTP server
        """
        self.logger.debug(f'Shutting down server')
        await super().shutdown()
        self.server.should_exit = True
        await self.server.shutdown()
        self.logger.debug(f'Server shutdown finished')

    async def run_server(self):
        """Run HTTP server forever"""
        await self.server.serve()

    @property
    def _should_exit(self):
        """Property describing if server is ready to exit
        :return: boolean indicating if Server ready to exit
        """
        return self.server.should_exit

    @property
    def should_exit(self):
        """Property describing if server is ready to exit
        :return: boolean indicating if Server ready to exit
        """
        return self._should_exit

    @staticmethod
    def is_ready(
        ctrl_address: str, timeout: float = 1.0, logger=None, **kwargs
    ) -> bool:
        """
        Check if status is ready.
        :param ctrl_address: the address where the control request needs to be sent
        :param timeout: timeout of the health check in seconds
        :param logger: JinaLogger to be used
        :param kwargs: extra keyword arguments
        :return: True if status is ready else False.
        """
        import urllib
        from http import HTTPStatus

        try:
            conn = urllib.request.urlopen(url=f'http://{ctrl_address}', timeout=timeout)
            return conn.code == HTTPStatus.OK
        except Exception as exc:
            if logger:
                logger.debug(f'Exception: {exc}')

            return False

    @staticmethod
    async def async_is_ready(
        ctrl_address: str, timeout: float = 1.0, logger=None, **kwargs
    ) -> bool:
        """
        Async Check if status is ready.
        :param ctrl_address: the address where the control request needs to be sent
        :param timeout: timeout of the health check in seconds
        :param logger: JinaLogger to be used
        :param kwargs: extra keyword arguments
        :return: True if status is ready else False.
        """
        return FastAPIBaseServer.is_ready(ctrl_address, timeout, logger=logger)


def _install_health_check(app: 'FastAPI', logger):
    health_check_exists = False
    for route in app.routes:
        if getattr(route, 'path', None) == '/' and 'GET' in getattr(
            route, 'methods', []
        ):
            health_check_exists = True
            logger.warning(
                'endpoint GET on "/" is used for health checks, make sure it\'s still accessible'
            )

    if not health_check_exists:
        from jina.serve.runtimes.gateway.health_model import JinaHealthModel

        @app.get(
            path='/',
            summary='Get the health of Jina Gateway service',
            response_model=JinaHealthModel,
        )
        async def _gateway_health():
            """
            Get the health of this Gateway service.
            .. # noqa: DAR201

            """
            return {}


class HTTPServer(FastAPIBaseServer):
    """
    :class:`HTTPServer` is a FastAPIBaseServer that uses the default FastAPI app for a given request handler
    """

    @property
    def app(self):
        """Get the default base API app for Server
        :return: Return a FastAPI app for the default HTTPGateway
        """
        return self._request_handler._http_fastapi_default_app(
            title=self.title,
            description=self.description,
            no_crud_endpoints=self.no_crud_endpoints,
            no_debug_endpoints=self.no_debug_endpoints,
            expose_endpoints=self.expose_endpoints,
            expose_graphql_endpoint=self.expose_graphql_endpoint,
            tracing=self.tracing,
            tracer_provider=self.tracer_provider,
            cors=self.cors,
            logger=self.logger,
        )


class SagemakerHTTPServer(FastAPIBaseServer):
    """
    :class:`SagemakerHTTPServer` is a FastAPIBaseServer that uses a custom FastAPI app for sagemaker endpoints

    """

    @property
    def port(self):
        """Get the port for the sagemaker server
        :return: Return the port for the sagemaker server, always 8080"""
        return 8080

    @property
    def ports(self):
        """Get the port for the sagemaker server
        :return: Return the port for the sagemaker server, always 8080"""
        return [8080]

    @property
    def app(self):
        """Get the sagemaker fastapi app
        :return: Return a FastAPI app for the sagemaker container
        """
        return self._request_handler._http_fastapi_csp_app(
            title=self.title,
            description=self.description,
            no_crud_endpoints=self.no_crud_endpoints,
            no_debug_endpoints=self.no_debug_endpoints,
            expose_endpoints=self.expose_endpoints,
            expose_graphql_endpoint=self.expose_graphql_endpoint,
            tracing=self.tracing,
            tracer_provider=self.tracer_provider,
            cors=self.cors,
            logger=self.logger,
        )


class AzureHTTPServer(FastAPIBaseServer):
    """
    :class:`AzureHTTPServer` is a FastAPIBaseServer that uses a custom FastAPI app for azure endpoints

    """

    @property
    def port(self):
        """Get the port for the azure server
        :return: Return the port for the azure server, always 8080"""
        return 8080

    @property
    def ports(self):
        """Get the port for the azure server
        :return: Return the port for the azure server, always 8080"""
        return [8080]

    @property
    def app(self):
        """Get the azure fastapi app
        :return: Return a FastAPI app for the azure container
        """
        return self._request_handler._http_fastapi_csp_app(
            title=self.title,
            description=self.description,
            no_crud_endpoints=self.no_crud_endpoints,
            no_debug_endpoints=self.no_debug_endpoints,
            expose_endpoints=self.expose_endpoints,
            expose_graphql_endpoint=self.expose_graphql_endpoint,
            tracing=self.tracing,
            tracer_provider=self.tracer_provider,
            cors=self.cors,
            logger=self.logger,
        )



import argparse
import asyncio
import signal
import time
from typing import TYPE_CHECKING, Optional, Union

from jina.constants import __windows__
from jina.excepts import PortAlreadyUsed, RuntimeTerminated
from jina.helper import ArgNamespace, is_port_free, random_ports, send_telemetry_event
from jina.logging.logger import JinaLogger
from jina.parsers import set_gateway_parser
from jina.parsers.helper import _set_gateway_uses

# Keep these imports even if not used, since YAML parser needs to find them in imported modules
from jina.serve.runtimes.gateway.composite import CompositeGateway
from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.gateway.grpc import GRPCGateway
from jina.serve.runtimes.gateway.http import HTTPGateway
from jina.serve.runtimes.gateway.websocket import WebSocketGateway
from jina.serve.runtimes.servers import BaseServer

if TYPE_CHECKING:  # pragma: no cover
    import threading
    import multiprocessing

HANDLED_SIGNALS = (
    signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
    signal.SIGTERM,  # Unix signal 15. Sent by `kill <pid>`.
    signal.SIGSEGV,
)


class AsyncNewLoopRuntime:
    """
    Runtime to make sure that a server can asynchronously run inside a new asynchronous loop. It will make sure that the server is run forever while handling the TERMINATE signals
    to be received by the orchestrator to shutdown the server and its resources.
    """

    def __init__(
        self,
        args: 'argparse.Namespace',
        cancel_event: Optional[
            Union['asyncio.Event', 'multiprocessing.Event', 'threading.Event']
        ] = None,
        signal_handlers_installed_event: Optional[
            Union['asyncio.Event', 'multiprocessing.Event', 'threading.Event']
        ] = None,
        req_handler_cls=None,
        gateway_load_balancer: bool = False,
        **kwargs,
    ):
        self.req_handler_cls = req_handler_cls
        self.gateway_load_balancer = gateway_load_balancer
        self.args = args
        if args.name:
            self.name = f'{args.name}/{self.__class__.__name__}'
        else:
            self.name = self.__class__.__name__
        self.logger = JinaLogger(self.name, **vars(self.args))
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)
        self.is_cancel = cancel_event or asyncio.Event()
        self.is_signal_handlers_installed = (
            signal_handlers_installed_event or asyncio.Event()
        )

        self.logger.debug(f'Setting signal handlers')

        if not __windows__:

            def _cancel(sig):
                def _inner_cancel(*args, **kwargs):
                    self.logger.debug(f'Received signal {sig.name}')
                    self.is_cancel.set(),

                return _inner_cancel

            for sig in HANDLED_SIGNALS:
                self._loop.add_signal_handler(sig, _cancel(sig), sig, None)
        else:

            def _cancel(signum, frame):
                self.logger.debug(f'Received signal {signum}')
                self.is_cancel.set(),

            for sig in HANDLED_SIGNALS:
                signal.signal(sig, _cancel)

        self.logger.debug(f'Signal handlers already set')
        self.is_signal_handlers_installed.set()
        self._start_time = time.time()
        self._loop.run_until_complete(self.async_setup())
        self._send_telemetry_event(event='start')

    def run_forever(self):
        """
        Running method to block the main thread.

        Run the event loop until a Future is done.
        """
        self._loop.run_until_complete(self._loop_body())

    def teardown(self):
        """Call async_teardown() and stop and close the event loop."""
        self._loop.run_until_complete(self.async_teardown())
        self._loop.stop()
        self._loop.close()
        self.logger.close()
        self._stop_time = time.time()
        self._send_telemetry_event(
            event='stop', extra_kwargs={'duration': self._stop_time - self._start_time}
        )

    async def _wait_for_cancel(self):
        """Do NOT override this method when inheriting from :class:`GatewayPod`"""
        # threads are not using asyncio.Event, but threading.Event
        if isinstance(self.is_cancel, asyncio.Event) and not hasattr(
            self.server, '_should_exit'
        ):
            await self.is_cancel.wait()
        else:
            while not self.is_cancel.is_set() and not getattr(
                self.server, '_should_exit', False
            ):
                await asyncio.sleep(0.1)

        await self.async_teardown()

    async def _loop_body(self):
        """Do NOT override this method when inheriting from :class:`GatewayPod`"""
        try:
            await asyncio.gather(self.async_run_forever(), self._wait_for_cancel())
        except asyncio.CancelledError:
            self.logger.warning('received terminate ctrl message from main process')

    def _cancel(self):
        """
        Signal the runtime to terminate
        """
        self.is_cancel.set()

    def _get_server(self):
        # construct server type based on protocol (and potentially req handler class to keep backwards compatibility)
        from jina.enums import ProtocolType, ProviderType

        if self.req_handler_cls.__name__ == 'GatewayRequestHandler':
            self.timeout_send = self.args.timeout_send
            if self.timeout_send:
                self.timeout_send /= 1e3  # convert ms to seconds
            if not self.args.port:
                self.args.port = random_ports(len(self.args.protocol))
            _set_gateway_uses(
                self.args, gateway_load_balancer=self.gateway_load_balancer
            )
            uses_with = self.args.uses_with or {}
            non_defaults = ArgNamespace.get_non_defaults_args(
                self.args, set_gateway_parser()
            )
            if 'title' not in non_defaults:
                uses_with['title'] = self.args.title
            if 'description' not in non_defaults:
                uses_with['description'] = self.args.description
            if 'no_debug_endpoints' not in non_defaults:
                uses_with['no_debug_endpoints'] = self.args.no_debug_endpoints
            if 'no_crud_endpoints' not in non_defaults:
                uses_with['no_crud_endpoints'] = self.args.no_crud_endpoints
            if 'expose_endpoints' not in non_defaults:
                uses_with['expose_endpoints'] = self.args.expose_endpoints
            if 'expose_graphql_endpoint' not in non_defaults:
                uses_with['expose_graphql_endpoint'] = self.args.expose_graphql_endpoint
            if 'cors' not in non_defaults:
                uses_with['cors'] = self.args.cors
            server = BaseGateway.load_config(
                self.args.uses,
                uses_with=dict(
                    **non_defaults,
                    **uses_with,
                ),
                uses_metas={},
                runtime_args={  # these are not parsed to the yaml config file but are pass directly during init
                    **vars(self.args),
                    'default_port': getattr(self.args, 'default_port', False),
                    'gateway_load_balancer': self.gateway_load_balancer,
                    'timeout_send': self.timeout_send,
                },
                py_modules=self.args.py_modules,
                extra_search_paths=self.args.extra_search_paths,
            )
            if isinstance(server, BaseServer):
                server.is_cancel = self.is_cancel
            return server
        elif (
            hasattr(self.args, 'provider')
            and self.args.provider == ProviderType.SAGEMAKER
        ):
            from jina.serve.runtimes.servers.http import SagemakerHTTPServer

            return SagemakerHTTPServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                proxy=getattr(self.args, 'proxy', None),
                uvicorn_kwargs=getattr(self.args, 'uvicorn_kwargs', None),
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                cors=getattr(self.args, 'cors', None),
                is_cancel=self.is_cancel,
            )
        elif (
            hasattr(self.args, 'provider')
            and self.args.provider == ProviderType.AZURE
        ):
            from jina.serve.runtimes.servers.http import AzureHTTPServer

            return AzureHTTPServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                proxy=getattr(self.args, 'proxy', None),
                uvicorn_kwargs=getattr(self.args, 'uvicorn_kwargs', None),
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                cors=getattr(self.args, 'cors', None),
                is_cancel=self.is_cancel,
            )
        elif not hasattr(self.args, 'protocol') or (
            len(self.args.protocol) == 1 and self.args.protocol[0] == ProtocolType.GRPC
        ):
            from jina.serve.runtimes.servers.grpc import GRPCServer

            return GRPCServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                grpc_server_options=self.args.grpc_server_options,
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                proxy=getattr(self.args, 'proxy', None),
            )

        elif (
            len(self.args.protocol) == 1 and self.args.protocol[0] == ProtocolType.HTTP
        ):
            from jina.serve.runtimes.servers.http import (
                HTTPServer,  # we need a concrete implementation of this
            )

            return HTTPServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                proxy=getattr(self.args, 'proxy', None),
                uvicorn_kwargs=getattr(self.args, 'uvicorn_kwargs', None),
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                cors=getattr(self.args, 'cors', None),
                is_cancel=self.is_cancel,
            )
        elif (
            len(self.args.protocol) == 1
            and self.args.protocol[0] == ProtocolType.WEBSOCKET
        ):
            from jina.serve.runtimes.servers.websocket import (
                WebSocketServer,  # we need a concrete implementation of this
            )

            return WebSocketServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                proxy=getattr(self.args, 'proxy', None),
                uvicorn_kwargs=getattr(self.args, 'uvicorn_kwargs', None),
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                is_cancel=self.is_cancel,
            )
        elif len(self.args.protocol) > 1:
            from jina.serve.runtimes.servers.composite import (
                CompositeServer,  # we need a concrete implementation of this
            )

            return CompositeServer(
                name=self.args.name,
                runtime_args=self.args,
                req_handler_cls=self.req_handler_cls,
                ssl_keyfile=getattr(self.args, 'ssl_keyfile', None),
                ssl_certfile=getattr(self.args, 'ssl_certfile', None),
                is_cancel=self.is_cancel,
            )

    def _send_telemetry_event(self, event, extra_kwargs=None):
        gateway_kwargs = {}
        if self.req_handler_cls.__name__ == 'WorkerRequestHandler':
            runtime_cls_name = 'WorkerRuntime'
        elif self.req_handler_cls.__name__ == 'HeaderRequestHandler':
            runtime_cls_name = 'HeadRuntime'
        else:
            runtime_cls_name = self.server.__class__.__name__
            gateway_kwargs['is_custom_gateway'] = self.server.__class__ not in [
                CompositeGateway,
                GRPCGateway,
                HTTPGateway,
                WebSocketGateway,
            ]
            gateway_kwargs['protocol'] = self.args.protocol

        extra_kwargs = extra_kwargs or {}

        send_telemetry_event(
            event=event,
            obj_cls_name=runtime_cls_name,
            entity_id=self._entity_id,
            **gateway_kwargs,
            **extra_kwargs,
        )

    async def async_setup(self):
        """
        The async method setup the runtime.

        Setup the uvicorn server.
        """
        if not (is_port_free(self.args.host, self.args.port)):
            raise PortAlreadyUsed(f'port:{self.args.port}')

        self.server = self._get_server()
        await self.server.setup_server()

    async def async_teardown(self):
        """Shutdown the server."""
        await self.server.shutdown()

    async def async_run_forever(self):
        """Running method of the server."""
        await self.server.run_server()
        self.is_cancel.set()

    @property
    def _entity_id(self):
        import uuid

        if hasattr(self, '_entity_id_'):
            return self._entity_id_
        self._entity_id_ = uuid.uuid1().hex
        return self._entity_id_

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type == RuntimeTerminated:
            self.logger.debug(f'{self!r} is ended')
        elif exc_type == KeyboardInterrupt:
            self.logger.debug(f'{self!r} is interrupted by user')
        elif exc_type and issubclass(exc_type, Exception):
            self.logger.error(
                (
                    f'{exc_val!r} during {self.run_forever!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else ''
                ),
                exc_info=not self.args.quiet_error,
            )
        try:
            self.teardown()
        except OSError:
            # OSError(Stream is closed) already
            pass
        except Exception as ex:
            self.logger.error(
                (
                    f'{ex!r} during {self.teardown!r}'
                    + f'\n add "--quiet-error" to suppress the exception details'
                    if not self.args.quiet_error
                    else ''
                ),
                exc_info=not self.args.quiet_error,
            )

        # https://stackoverflow.com/a/28158006
        # return True will silent all exception stack trace here, silence is desired here as otherwise it is too
        # noisy
        #
        # doc: If an exception is supplied, and the method wishes to suppress the exception (i.e., prevent it
        # from being propagated), it should return a true value. Otherwise, the exception will be processed normally
        # upon exit from this method.
        return True



from typing import TYPE_CHECKING, Dict, List, Optional, Union

from jina.excepts import InternalNetworkError
from jina.helper import get_full_version
from jina.importer import ImportExtensions
from jina.logging.logger import JinaLogger
from jina.serve.networking.sse import EventSourceResponse
from jina.types.request.data import DataRequest

if TYPE_CHECKING:  # pragma: no cover
    from opentelemetry import trace

    from jina.serve.runtimes.gateway.streamer import GatewayStreamer


def get_fastapi_app(
    streamer: 'GatewayStreamer',
    title: str,
    description: str,
    expose_graphql_endpoint: bool,
    cors: bool,
    logger: 'JinaLogger',
    tracing: Optional[bool] = None,
    tracer_provider: Optional['trace.TracerProvider'] = None,
    **kwargs,
):
    """
    Get the app from FastAPI as the REST interface.

    :param streamer: gateway streamer object
    :param title: The title of this HTTP server. It will be used in automatics docs such as Swagger UI.
    :param description: The description of this HTTP server. It will be used in automatics docs such as Swagger UI.
    :param expose_graphql_endpoint: If set, /graphql endpoint is added to HTTP interface.
    :param cors: If set, a CORS middleware is added to FastAPI frontend to allow cross-origin access.
    :param logger: Jina logger.
    :param tracing: Enables tracing if set to True.
    :param tracer_provider: If tracing is enabled the tracer_provider will be used to instrument the code.
    :param kwargs: Extra kwargs to make it compatible with other methods
    :return: fastapi app
    """
    if expose_graphql_endpoint:
        logger.error(f' GraphQL endpoint is not enabled when using docarray >0.30')
    with ImportExtensions(required=True):
        from fastapi import FastAPI, Response, HTTPException
        from fastapi.middleware.cors import CORSMiddleware
        import pydantic
        from pydantic import Field
    from docarray import BaseDoc, DocList
    from docarray.base_doc.docarray_response import DocArrayResponse

    from jina import __version__

    app = FastAPI(
        title=title or 'My Jina Service',
        description=description
        or 'This is my awesome service. You can set `title` and `description` in your `Flow` or `Gateway` '
        'to customize the title and description.',
        version=__version__,
    )

    if tracing:
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        FastAPIInstrumentor.instrument_app(app, tracer_provider=tracer_provider)

    if cors:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*'],
        )
        logger.warning('CORS is enabled. This service is accessible from any website!')

    @app.on_event('shutdown')
    async def _shutdown():
        await streamer.close()

    import os

    from pydantic import BaseModel
    from pydantic.config import BaseConfig, inherit_config

    from jina.proto import jina_pb2
    from jina.serve.runtimes.gateway.models import (
        PROTO_TO_PYDANTIC_MODELS,
        _to_camel_case,
    )
    from jina.types.request.status import StatusMessage

    class Header(BaseModel):
        request_id: Optional[str] = Field(
            description='Request ID', example=os.urandom(16).hex()
        )
        target_executor: Optional[str] = Field(default=None, example="")

        class Config(BaseConfig):
            alias_generator = _to_camel_case
            allow_population_by_field_name = True

    class InnerConfig(BaseConfig):
        alias_generator = _to_camel_case
        allow_population_by_field_name = True

    @app.get(
        path='/dry_run',
        summary='Get the readiness of Jina Flow service, sends an empty DocumentArray to the complete Flow to '
        'validate connectivity',
        response_model=PROTO_TO_PYDANTIC_MODELS.StatusProto,
    )
    async def _flow_health():
        """
        Get the health of the complete Flow service.
        .. # noqa: DAR201

        """

        docs = DocList[BaseDoc]([])

        try:
            async for _ in streamer.stream_docs(docs, request_size=1):
                status_message = StatusMessage()
                status_message.set_code(jina_pb2.StatusProto.SUCCESS)
                return status_message.to_dict()
        except Exception as ex:
            status_message = StatusMessage()
            status_message.set_exception(ex)
            return status_message.to_dict(use_integers_for_enums=True)

    request_models_map = streamer._endpoints_models_map

    if '/status' not in request_models_map:
        from jina.serve.runtimes.gateway.health_model import JinaInfoModel

        @app.get(
            path='/status',
            summary='Get the status of Jina service',
            response_model=JinaInfoModel,
            tags=['Debug'],
        )
        async def _status():
            """
            Get the status of this Jina service.

            This is equivalent to running `jina -vf` from command line.

            .. # noqa: DAR201
            """
            version, env_info = get_full_version()
            for k, v in version.items():
                version[k] = str(v)
            for k, v in env_info.items():
                env_info[k] = str(v)
            return {'jina': version, 'envs': env_info}

    def _generate_exception_header(error: InternalNetworkError):
        import traceback

        exception_dict = {
            'name': str(error.__class__),
            'stacks': [
                str(x) for x in traceback.extract_tb(error.og_exception.__traceback__)
            ],
            'executor': '',
        }
        status_dict = {
            'code': DataRequest().status.ERROR,
            'description': error.details() if error.details() else '',
            'exception': exception_dict,
        }
        header_dict = {'request_id': error.request_id, 'status': status_dict}
        return header_dict

    def add_post_route(
        endpoint_path,
        input_model,
        output_model,
        input_doc_list_model=None,
        output_doc_list_model=None,
    ):
        app_kwargs = dict(
            path=f'/{endpoint_path.strip("/")}',
            methods=['POST'],
            summary=f'Endpoint {endpoint_path}',
            response_model=output_model,
        )
        app_kwargs['response_class'] = DocArrayResponse

        @app.api_route(**app_kwargs)
        async def post(body: input_model, response: Response):
            target_executor = None
            req_id = None
            if body.header is not None:
                target_executor = body.header.target_executor
                req_id = body.header.request_id
            data = body.data
            if isinstance(data, list):
                docs = DocList[input_doc_list_model](data)
            else:
                docs = DocList[input_doc_list_model]([data])
                if body.header is None:
                    if hasattr(docs[0], 'id'):
                        req_id = docs[0].id

            try:
                async for resp in streamer.stream_docs(
                    docs,
                    exec_endpoint=endpoint_path,
                    parameters=body.parameters,
                    target_executor=target_executor,
                    request_id=req_id,
                    return_results=True,
                    return_type=DocList[output_doc_list_model],
                ):
                    status = resp.header.status

                    if status.code == jina_pb2.StatusProto.ERROR:
                        raise HTTPException(status_code=499, detail=status.description)
                    else:
                        result_dict = resp.to_dict()
                        return result_dict
            except InternalNetworkError as err:
                import grpc

                if (
                    err.code() == grpc.StatusCode.UNAVAILABLE
                    or err.code() == grpc.StatusCode.NOT_FOUND
                ):
                    response.status_code = status.HTTP_503_SERVICE_UNAVAILABLE
                elif err.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                    response.status_code = status.HTTP_504_GATEWAY_TIMEOUT
                else:
                    response.status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
                result = body.dict()  # send back the request
                result['header'] = _generate_exception_header(
                    err
                )  # attach exception details to response header
                logger.error(
                    f'Error while getting responses from deployments: {err.details()}'
                )
                return result

    def add_streaming_routes(
        endpoint_path,
        input_doc_model=None,
    ):
        from fastapi import Request

        @app.api_route(
            path=f'/{endpoint_path.strip("/")}',
            methods=['GET'],
            summary=f'Streaming Endpoint {endpoint_path}',
        )
        async def streaming_get(request: Request, body: input_doc_model = None):
            body = body or dict(request.query_params)
            body = input_doc_model.parse_obj(body)

            async def event_generator():
                async for doc, error in streamer.stream_doc(
                    doc=body, exec_endpoint=endpoint_path
                ):
                    if error:
                        raise HTTPException(status_code=499, detail=str(error))
                    yield {'event': 'update', 'data': doc.dict()}
                yield {'event': 'end'}

            return EventSourceResponse(event_generator())

    for endpoint, input_output_map in request_models_map.items():
        if endpoint != '_jina_dry_run_':
            input_doc_model = input_output_map['input']
            output_doc_model = input_output_map['output']
            is_generator = input_output_map['is_generator']
            parameters_model = input_output_map['parameters']
            parameters_model_needed = parameters_model is not None
            if parameters_model_needed:
                try:
                    _ = parameters_model()
                    parameters_model_needed = False
                except:
                    parameters_model_needed = True
                parameters_model = parameters_model if parameters_model_needed else Optional[parameters_model]
                default_parameters = (
                    ... if parameters_model_needed else None
                )
            else:
                parameters_model = Optional[Dict]
                default_parameters = None

            _config = inherit_config(InnerConfig, BaseDoc.__config__)

            endpoint_input_model = pydantic.create_model(
                f'{endpoint.strip("/")}_input_model',
                data=(Union[List[input_doc_model], input_doc_model], ...),
                parameters=(parameters_model, default_parameters),
                header=(Optional[Header], None),
                __config__=_config,
            )

            endpoint_output_model = pydantic.create_model(
                f'{endpoint.strip("/")}_output_model',
                data=(Union[List[output_doc_model], output_doc_model], ...),
                parameters=(Optional[Dict], None),
                header=(Optional[Header], None),
                __config__=_config,
            )

            if is_generator:
                add_streaming_routes(
                    endpoint,
                    input_doc_model=input_doc_model,
                )
            else:
                add_post_route(
                    endpoint,
                    input_model=endpoint_input_model,
                    output_model=endpoint_output_model,
                    input_doc_list_model=input_doc_model,
                    output_doc_list_model=output_doc_model,
                )

    return app



from pydantic import BaseModel
from typing import Dict


def _to_camel_case(snake_str: str) -> str:
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return components[0] + ''.join(x.title() for x in components[1:])


class JinaHealthModel(BaseModel):
    """Pydantic BaseModel for Jina health check, used as the response model in REST app."""

    ...


class JinaInfoModel(BaseModel):
    """Pydantic BaseModel for Jina status, used as the response model in REST app."""

    jina: Dict
    envs: Dict

    class Config:
        alias_generator = _to_camel_case
        allow_population_by_field_name = True



from jina import Client
from docarray import DocList
from docarray.documents import TextDoc

if __name__ == '__main__':
    c = Client(host='grpc://0.0.0.0:54321')
    da = c.post(
        '/', DocList[TextDoc]([TextDoc(), TextDoc()]), return_type=DocList[TextDoc]
    )
    print(da.text)



from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.servers.grpc import GRPCServer

__all__ = ['GRPCGateway']


class GRPCGateway(GRPCServer, BaseGateway):
    """
    :class:`GRPCGateway` is a GRPCServer that can be loaded from YAML as any other Gateway
    """

    pass



from typing import Any, Dict, Optional, Type

from jina.jaml.parsers.base import BaseLegacyParser
from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.gateway.request_handling import GatewayRequestHandler


class GatewayLegacyParser(BaseLegacyParser):
    """Legacy parser for gateway."""

    def parse(
        self,
        cls: Type['BaseGateway'],
        data: Dict,
        runtime_args: Optional[Dict[str, Any]] = None,
    ) -> 'BaseGateway':
        """
        :param cls: target class type to parse into, must be a :class:`JAMLCompatible` type
        :param data: gateway yaml file loaded as python dict
        :param runtime_args: Optional runtime_args to be directly passed without being parsed into a yaml config
        :return: the Gateway YAML parser given the syntax version number
        """

        data['metas'] = {}

        cls._init_from_yaml = True
        # tmp_p = {kk: expand_env_var(vv) for kk, vv in data.get('with', {}).items()}

        for key in {
            'name',
            'port',
            'protocol',
            'host',
            'tracing',
            'graph_description',
            'graph_conditions',
            'deployments_addresses',
            'deployments_metadata',
            'deployments_no_reduce',
            'timeout_send',
            'retries',
            'compression',
            'runtime_name',
            'prefetch',
            'meter',
            'log_config',
        }:
            if runtime_args and not runtime_args.get(key) and data.get(key):
                runtime_args[key] = data.get(key)
        if runtime_args.get('default_port'):
            yaml_port = data.get('port')
            if isinstance(yaml_port, int):
                yaml_port = [yaml_port]
            runtime_args['port'] = yaml_port or runtime_args.get('port')

        obj = cls(
            **data.get('with', {}),
            metas=data.get('metas', {}),
            requests=data.get('requests', {}),
            runtime_args=runtime_args,
            req_handler_cls=GatewayRequestHandler
        )

        cls._init_from_yaml = False
        obj.is_updated = False
        return obj

    def dump(self, data: 'BaseGateway') -> Dict:
        """
        :param data: versioned gateway object
        :return: the dictionary given a versioned gateway object
        """
        a = {k: v for k, v in data._init_kwargs_dict.items()}
        r = {}
        if a:
            r['with'] = a

        return r



"""A module for the websockets-based Client for Jina."""

import asyncio
from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Dict, Optional, Tuple

from starlette import status

from jina.clients.base import BaseClient
from jina.clients.base.helper import WebsocketClientlet
from jina.clients.helper import callback_exec
from jina.helper import get_or_reuse_loop
from jina.importer import ImportExtensions
from jina.logging.profile import ProgressBar
from jina.proto import jina_pb2
from jina.serve.stream import RequestStreamer

if TYPE_CHECKING:  # pragma: no cover
    from jina.clients.base import CallbackFnType, InputType
    from jina.types.request import Request


class WebSocketBaseClient(BaseClient):
    """A Websocket Client."""

    async def _is_flow_ready(self, **kwargs) -> bool:
        """Sends a dry run to the Flow to validate if the Flow is ready to receive requests

        :param kwargs: kwargs coming from the public interface. Includes arguments to be passed to the `WebsocketClientlet`
        :return: boolean indicating the readiness of the Flow
        """
        async with AsyncExitStack() as stack:
            try:
                proto = 'wss' if self.args.tls else 'ws'
                url = f'{proto}://{self.args.host}:{self.args.port}/dry_run'
                iolet = await stack.enter_async_context(
                    WebsocketClientlet(
                        url=url,
                        logger=self.logger,
                        tracer_provider=self.tracer_provider,
                        **kwargs,
                    )
                )

                async def _receive():
                    try:
                        async for response in iolet.recv_dry_run():
                            return response
                    except Exception as exc:
                        self.logger.error(
                            f'Error while fetching response from Websocket server {exc!r}'
                        )
                        raise

                async def _send():
                    return await iolet.send_dry_run()

                receive_task = asyncio.create_task(_receive())

                if receive_task.done():
                    raise RuntimeError(
                        'receive task not running, can not send messages'
                    )
                try:
                    send_task = asyncio.create_task(_send())
                    _, response_result = await asyncio.gather(send_task, receive_task)
                    if response_result.proto.code == jina_pb2.StatusProto.SUCCESS:
                        return True
                finally:
                    if iolet.close_code == status.WS_1011_INTERNAL_ERROR:
                        raise ConnectionError(iolet.close_message)
                    await receive_task

            except Exception as e:
                self.logger.error(
                    f'Error while getting response from websocket server {e!r}'
                )
            return False

    async def _get_results(
        self,
        inputs: 'InputType',
        on_done: 'CallbackFnType',
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        max_attempts: int = 1,
        initial_backoff: float = 0.5,
        max_backoff: float = 0.1,
        backoff_multiplier: float = 1.5,
        results_in_order: bool = False,
        prefetch: Optional[int] = None,
        **kwargs,
    ):
        """
        :param inputs: the callable
        :param on_done: the callback for on_done
        :param on_error: the callback for on_error
        :param on_always: the callback for on_always
        :param max_attempts: Number of sending attempts, including the original request.
        :param initial_backoff: The first retry will happen with a delay of random(0, initial_backoff)
        :param max_backoff: The maximum accepted backoff after the exponential incremental delay
        :param backoff_multiplier: The n-th attempt will occur at random(0, min(initialBackoff*backoffMultiplier**(n-1), maxBackoff))
        :param results_in_order: return the results in the same order as the inputs
        :param prefetch: How many Requests are processed from the Client at the same time.
        :param kwargs: kwargs coming from the public interface. Includes arguments to be passed to the `WebsocketClientlet`
        :yields: generator over results
        """
        with ImportExtensions(required=True):
            pass

        self.inputs = inputs
        request_iterator = self._get_requests(**kwargs)

        async with AsyncExitStack() as stack:
            cm1 = ProgressBar(
                total_length=self._inputs_length, disable=not (self.show_progress)
            )
            p_bar = stack.enter_context(cm1)

            proto = 'wss' if self.args.tls else 'ws'
            url = f'{proto}://{self.args.host}:{self.args.port}/'
            iolet = await stack.enter_async_context(
                WebsocketClientlet(
                    url=url,
                    logger=self.logger,
                    tracer_provider=self.tracer_provider,
                    max_attempts=max_attempts,
                    initial_backoff=initial_backoff,
                    max_backoff=max_backoff,
                    backoff_multiplier=backoff_multiplier,
                    **kwargs,
                )
            )

            request_buffer: Dict[str, asyncio.Future] = (
                dict()
            )  # maps request_ids to futures (tasks)

            def _result_handler(result):
                return result

            async def _receive():
                def _response_handler(response):
                    if response.header.request_id in request_buffer:
                        future = request_buffer.pop(response.header.request_id)
                        future.set_result(response)
                    else:
                        self.logger.warning(
                            f'discarding unexpected response with request id {response.header.request_id}'
                        )

                """Await messages from WebsocketGateway and process them in the request buffer"""
                try:
                    async for response in iolet.recv_message():
                        _response_handler(response)
                finally:
                    if request_buffer:
                        self.logger.warning(
                            f'{self.__class__.__name__} closed, cancelling all outstanding requests'
                        )
                        for future in request_buffer.values():
                            future.cancel()
                        request_buffer.clear()

            def _handle_end_of_iter():
                """Send End of iteration signal to the Gateway"""
                asyncio.create_task(iolet.send_eoi())

            def _request_handler(
                request: 'Request', **kwargs
            ) -> 'Tuple[asyncio.Future, Optional[asyncio.Future]]':
                """
                For each request in the iterator, we send the `Message` using `iolet.send_message()`.
                For websocket requests from client, for each request in the iterator, we send the request in `bytes`
                using `iolet.send_message()`.
                Then add {<request-id>: <an-empty-future>} to the request buffer.
                This empty future is used to track the `result` of this request during `receive`.
                :param request: current request in the iterator
                :param kwargs: kwargs
                :return: asyncio Future for sending message
                """
                future = get_or_reuse_loop().create_future()
                request_buffer[request.header.request_id] = future
                asyncio.create_task(iolet.send_message(request))
                return future, None

            streamer_args = vars(self.args)
            if prefetch:
                streamer_args['prefetch'] = prefetch
            streamer = RequestStreamer(
                request_handler=_request_handler,
                result_handler=_result_handler,
                end_of_iter_handler=_handle_end_of_iter,
                logger=self.logger,
                **streamer_args,
            )

            receive_task = asyncio.create_task(_receive())

            exception_raised = None

            if receive_task.done():
                raise RuntimeError('receive task not running, can not send messages')
            try:
                async for response in streamer.stream(
                    request_iterator=request_iterator,
                    results_in_order=results_in_order,
                ):
                    callback_exec(
                        response=response,
                        logger=self.logger,
                        on_error=on_error,
                        on_done=on_done,
                        on_always=on_always,
                        continue_on_error=self.continue_on_error,
                    )
                    if self.show_progress:
                        p_bar.update()
                    yield response
            except Exception as ex:
                exception_raised = ex
                try:
                    receive_task.cancel()
                except:
                    raise ex
            finally:
                if iolet.close_code == status.WS_1011_INTERNAL_ERROR:
                    raise ConnectionError(iolet.close_message)
                try:
                    await receive_task
                except asyncio.CancelledError:
                    if exception_raised is not None:
                        raise exception_raised
                    else:
                        raise



import os
import re

from typing import TYPE_CHECKING, Tuple, Dict, Optional, cast
from pkg_resources import Requirement

if TYPE_CHECKING:  # pragma: no cover
    from pathlib import Path

VCS_SCHEMES = [
    'git',
    'git+https',
    'git+ssh',
    'git+git',
    'hg+http',
    'hg+https',
    'hg+static-http',
    'hg+ssh',
    'svn',
    'svn+svn',
    'svn+http',
    'svn+https',
    'svn+ssh',
    'bzr+http',
    'bzr+https',
    'bzr+ssh',
    'bzr+sftp',
    'bzr+ftp',
    'bzr+lp',
]

URI_REGEX = re.compile(
    r'^(?P<scheme>https?|file|ftps?)://(?P<path>[^#]+)' r'(#(?P<fragment>\S+))?'
)

VCS_SCHEMES_REGEX = r'|'.join([scheme.replace('+', r'\+') for scheme in VCS_SCHEMES])
VCS_REGEX = re.compile(
    rf'^(?P<scheme>{VCS_SCHEMES_REGEX})://((?P<login>[^/@]+)@)?'
    r'(?P<path>[^#@]+)(@(?P<revision>[^#]+))?(#(?P<fragment>\S+))?'
)
ENV_VAR_RE = re.compile(r"(?P<var>\$\{(?P<name>[A-Z0-9_]+)\})")

ENV_VAR_RE_ONLY_MATCH_UPPERCASE_UNDERLINE = re.compile(r"^[A-Z0-9_]+$")

extras_require_search = re.compile(r'(?P<name>.+)\[(?P<extras>[^\]]+)\]')


def _parse_fragment(fragment_string: str) -> Dict[str, str]:
    """Takes a fragment string and returns a dict of the components

    :param fragment_string: a fragment string
    :return: a dict of components
    """
    fragment_string = fragment_string.lstrip('#')

    try:
        return dict(
            cast(Tuple[str, str], tuple(key_value_string.split('=')))
            for key_value_string in fragment_string.split('&')
        )
    except ValueError:
        raise ValueError(f'Invalid fragment string {fragment_string}')


def _parse_requirement(line: str) -> 'Requirement':
    """Parses a Requirement from a line of a requirement file.

    :param line: a line of a requirement file
    :returns: a Requirement instance for the given line
    """
    vcs_match = VCS_REGEX.match(line)
    uri_match = URI_REGEX.match(line)

    if vcs_match is not None:
        groups = vcs_match.groupdict()
        name = os.path.basename(groups['path']).split('.')[0]
        egg = None
        if groups['fragment']:
            fragment = _parse_fragment(groups['fragment'])
            egg = fragment.get('egg')

        line = f'{egg or name} @ {line}'
    elif uri_match is not None:
        groups = uri_match.groupdict()
        name = os.path.basename(groups['path']).split('.')[0]
        egg = None
        if groups['fragment']:
            fragment = _parse_fragment(groups['fragment'])
            egg = fragment.get('egg')

        line = f'{egg or name} @ {line}'

    return Requirement.parse(line)


def _expand_env_variables(line: str) -> str:
    """
    Replace all environment variables that can be retrieved via `os.getenv`.
    The only allowed format for environment variables defined in the
    requirement file is `${MY_VARIABLE_1}` to ensure two things:
    1. Strings that contain a `$` aren't accidentally (partially) expanded.
    2. Ensure consistency across platforms for requirement files.
    Valid characters in variable names follow the `POSIX standard
    <http://pubs.opengroup.org/onlinepubs/9699919799/>`_ and are limited
    to uppercase letter and number and the `_` (underscore).
    Replace environment variables in requirement if it's defined.
    :param line: a line of a requirement file
    :return: line
    """
    for env_var, var_name in ENV_VAR_RE.findall(line):
        value = os.getenv(var_name)
        if not value:
            raise Exception(
                f'The given requirements.txt require environment variables `{var_name}` does not exist!'
            )
        line = line.replace(env_var, value)
    return line


def _get_install_options(requirements_file: 'Path', excludes: Tuple[str] = ('jina',)):
    with requirements_file.open() as requirements:
        install_options = []
        install_reqs = []
        for req in requirements:
            req = req.strip()
            if (not req) or req.startswith('#'):
                continue
            elif req.startswith('-'):
                for index, item in enumerate(req.split(' ')):
                    install_options.append(_expand_env_variables(item))
            else:
                expand_req = _expand_env_variables(req)
                req_spec = _parse_requirement(expand_req)

                if req_spec.project_name not in excludes or len(req_spec.extras) > 0:
                    install_reqs.append(expand_req)
    return install_reqs, install_options


def _is_requirements_installed(requirements_file: 'Path') -> bool:
    """Return True if requirements.txt is installed locally
    :param requirements_file: the requirements.txt file
    :return: True or False if not satisfied
    """
    import pkg_resources
    from pkg_resources import (
        DistributionNotFound,
        RequirementParseError,
        VersionConflict,
    )

    install_reqs, install_options = _get_install_options(requirements_file)

    if len(install_reqs) == 0:
        return True

    try:
        pkg_resources.require('\n'.join(install_reqs))
    except (DistributionNotFound, VersionConflict, RequirementParseError) as ex:
        import warnings

        warnings.warn(repr(ex))
        return isinstance(ex, VersionConflict)
    return True


def _install_requirements(requirements_file: 'Path', timeout: int = 1000):
    """Install modules included in requirements file
    :param requirements_file: the requirements.txt file
    :param timeout: the socket timeout (default = 1000s)
    """
    import subprocess
    import sys

    if _is_requirements_installed(requirements_file):
        return

    install_reqs, install_options = _get_install_options(requirements_file)

    subprocess.check_call(
        [
            sys.executable,
            '-m',
            'pip',
            'install',
            '--compile',
            f'--default-timeout={timeout}',
        ]
        + install_reqs
        + install_options
    )


def install_package_dependencies(pkg_path: Optional['Path']) -> None:
    """

    :param pkg_path: package path
    """
    # install the dependencies included in requirements.txt
    if pkg_path:
        requirements_file = pkg_path / 'requirements.txt'

        if requirements_file.exists():
            _install_requirements(requirements_file)


def _get_package_path_from_uses(uses: str) -> Optional['Path']:
    if isinstance(uses, str) and os.path.exists(uses):
        from pathlib import Path

        return Path(os.path.dirname(os.path.abspath(uses)))
    else:
        from hubble.executor.helper import is_valid_huburi

        if not is_valid_huburi(uses):
            from jina.logging.predefined import default_logger

            default_logger.warning(
                f'Error getting the directory name from {uses}. `--install-requirements` option is only valid when `uses` is a configuration file.'
            )
        return None



from abc import ABC
from contextlib import ExitStack

from rich.table import Table

from jina.helper import CatchAllCleanupContextManager, get_internal_ip, get_public_ip


class BaseOrchestrator(ExitStack, ABC):
    """Base orchestrator class"""

    def __enter__(self):
        with CatchAllCleanupContextManager(self):
            return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if hasattr(self, '_stop_event'):
            self._stop_event.set()

        super().__exit__(exc_type, exc_val, exc_tb)

    def _init_table(self):
        table = Table(
            title=None, box=None, highlight=True, show_header=False, min_width=40
        )
        table.add_column('', justify='left')
        table.add_column('', justify='right')
        table.add_column('', justify='right')
        return table

    @property
    def address_private(self) -> str:
        """Return the private IP address of the gateway for connecting from other machine in the same network


        .. # noqa: DAR201"""
        if getattr(self, '_internal_ip', None):
            return self._internal_ip
        else:
            self._internal_ip = get_internal_ip()
        return self._internal_ip

    @property
    def address_public(self) -> str:
        """Return the public IP address of the gateway for connecting from other machine in the public network


        .. # noqa: DAR201"""
        if getattr(self, '_public_ip', None):
            return self._public_ip
        else:
            self._public_ip = get_public_ip()
        return self._public_ip

    @property
    def _entity_id(self) -> str:
        import uuid

        if hasattr(self, '_entity_id_'):
            return self._entity_id_
        self._entity_id_ = uuid.uuid1().hex
        return self._entity_id_



import itertools
from typing import TYPE_CHECKING, AsyncIterator, Dict

from jina.enums import ProtocolType
from jina.helper import get_full_version
from jina.proto import jina_pb2
from jina.types.request.data import DataRequest, SingleDocumentRequest
from jina.types.request.status import StatusMessage

if TYPE_CHECKING:  # pragma: no cover
    from types import SimpleNamespace

    import grpc

    from jina.logging.logger import JinaLogger
    from jina.serve.runtimes.gateway.streamer import GatewayStreamer
    from jina.types.request import Request


class GatewayRequestHandler:
    """Object to encapsulate the code related to handle the data requests in the Gateway"""

    def __init__(
        self,
        args: 'SimpleNamespace',
        logger: 'JinaLogger',
        metrics_registry=None,
        meter=None,
        aio_tracing_client_interceptors=None,
        tracing_client_interceptor=None,
        **kwargs,
    ):
        import json

        from jina.serve.runtimes.gateway.streamer import (
            GatewayStreamer,
            _ExecutorStreamer,
        )

        self.runtime_args = args
        self.logger = logger
        graph_description = json.loads(self.runtime_args.graph_description)
        graph_conditions = json.loads(self.runtime_args.graph_conditions)
        deployments_addresses = json.loads(self.runtime_args.deployments_addresses)
        deployments_metadata = json.loads(self.runtime_args.deployments_metadata)
        deployments_no_reduce = json.loads(self.runtime_args.deployments_no_reduce)

        deployment_grpc_addresses = {}
        for deployment_name, addresses in deployments_addresses.items():
            if isinstance(addresses, Dict):
                deployment_grpc_addresses[deployment_name] = addresses.get(
                    ProtocolType.GRPC.to_string(), []
                )
            else:
                deployment_grpc_addresses[deployment_name] = addresses

        self.streamer = GatewayStreamer(
            graph_representation=graph_description,
            executor_addresses=deployment_grpc_addresses,
            graph_conditions=graph_conditions,
            deployments_metadata=deployments_metadata,
            deployments_no_reduce=deployments_no_reduce,
            timeout_send=self.runtime_args.timeout_send,
            retries=self.runtime_args.retries,
            compression=self.runtime_args.compression,
            runtime_name=self.runtime_args.runtime_name,
            prefetch=self.runtime_args.prefetch,
            logger=self.logger,
            metrics_registry=metrics_registry,
            meter=meter,
            aio_tracing_client_interceptors=aio_tracing_client_interceptors,
            tracing_client_interceptor=tracing_client_interceptor,
            grpc_channel_options=(
                self.runtime_args.grpc_channel_options
                if hasattr(self.runtime_args, 'grpc_channel_options')
                else None
            ),
        )

        GatewayStreamer._set_env_streamer_args(
            graph_representation=graph_description,
            executor_addresses=deployment_grpc_addresses,
            graph_conditions=graph_conditions,
            deployments_metadata=deployments_metadata,
            deployments_no_reduce=deployments_no_reduce,
            timeout_send=self.runtime_args.timeout_send,
            retries=self.runtime_args.retries,
            compression=self.runtime_args.compression,
            runtime_name=self.runtime_args.runtime_name,
            prefetch=self.runtime_args.prefetch,
        )

        self.executor = {
            executor_name: _ExecutorStreamer(
                self.streamer._connection_pool, executor_name=executor_name
            )
            for executor_name in deployment_grpc_addresses.keys()
        }
        servers = []
        for addresses in deployments_addresses.values():
            if isinstance(addresses, Dict):
                servers.extend(addresses.get(ProtocolType.HTTP.to_string(), []))
        self.load_balancer_servers = itertools.cycle(servers)

    async def close(self):
        """
        Gratefully closes the object making sure all the floating requests are taken care and the connections are closed gracefully
        """
        self.logger.debug(f'Closing Request Handler')
        await self.streamer.close()
        self.logger.debug(f'Request Handler closed')

    def _http_fastapi_default_app(
        self,
        title,
        description,
        no_debug_endpoints,
        no_crud_endpoints,
        expose_endpoints,
        expose_graphql_endpoint,
        cors,
        tracing,
        tracer_provider,
        **kwargs,
    ):
        from jina._docarray import docarray_v2
        from jina.helper import extend_rest_interface

        if not docarray_v2:
            from jina.serve.runtimes.gateway.http_fastapi_app import get_fastapi_app
        else:
            from jina.serve.runtimes.gateway.http_fastapi_app_docarrayv2 import (
                get_fastapi_app,
            )

        return extend_rest_interface(
            get_fastapi_app(
                streamer=self.streamer,
                title=title,
                description=description,
                no_debug_endpoints=no_debug_endpoints,
                no_crud_endpoints=no_crud_endpoints,
                expose_endpoints=expose_endpoints,
                expose_graphql_endpoint=expose_graphql_endpoint,
                cors=cors,
                logger=self.logger,
                tracing=tracing,
                tracer_provider=tracer_provider,
            )
        )

    async def _load_balance(self, request):
        import aiohttp
        from aiohttp import web

        target_server = next(self.load_balancer_servers)
        target_url = f'{target_server}{request.path_qs}'

        try:
            async with aiohttp.ClientSession() as session:

                request_kwargs = {}
                try:
                    payload = await request.json()
                    if payload:
                        request_kwargs['json'] = payload
                except Exception:
                    self.logger.debug('No JSON payload found in request')

                async with session.request(
                    request.method,
                    url=target_url,
                    auto_decompress=False,
                    **request_kwargs,
                ) as response:
                    # Create a StreamResponse with the same headers and status as the target response
                    stream_response = web.StreamResponse(
                        status=response.status,
                        headers=response.headers,
                    )

                    # Prepare the response to send headers
                    await stream_response.prepare(request)

                    # Stream the response from the target server to the client
                    async for chunk in response.content.iter_any():
                        await stream_response.write(chunk)

                    # Close the stream response once all chunks are sent
                    await stream_response.write_eof()
                    return stream_response

        except aiohttp.ClientError as e:
            return web.Response(text=f'Error: {str(e)}', status=500)

    def _websocket_fastapi_default_app(self, tracing, tracer_provider):
        from jina.helper import extend_rest_interface
        from jina.serve.runtimes.gateway.websocket_fastapi_app import get_fastapi_app

        return extend_rest_interface(
            get_fastapi_app(
                streamer=self.streamer,
                logger=self.logger,
                tracing=tracing,
                tracer_provider=tracer_provider,
            )
        )

    async def dry_run(self, empty, context) -> jina_pb2.StatusProto:
        """
        Process the call requested by having a dry run call to every Executor in the graph

        :param empty: The service expects an empty protobuf message
        :param context: grpc context
        :returns: the response request
        """
        self.logger.debug('recv a dry_run request')
        from jina._docarray import Document, DocumentArray
        from jina.serve.executors import __dry_run_endpoint__

        da = DocumentArray([Document()])
        try:
            async for _ in self.streamer.stream_docs(
                docs=da, exec_endpoint=__dry_run_endpoint__, request_size=1
            ):
                pass
            status_message = StatusMessage()
            status_message.set_code(jina_pb2.StatusProto.SUCCESS)
            return status_message.proto
        except Exception as ex:
            status_message = StatusMessage()
            status_message.set_exception(ex)
            return status_message.proto

    async def _status(self, empty, context) -> jina_pb2.JinaInfoProto:
        """
        Process the the call requested and return the JinaInfo of the Runtime

        :param empty: The service expects an empty protobuf message
        :param context: grpc context
        :returns: the response request
        """
        self.logger.debug('recv a _status request')
        info_proto = jina_pb2.JinaInfoProto()
        version, env_info = get_full_version()
        for k, v in version.items():
            info_proto.jina[k] = str(v)
        for k, v in env_info.items():
            info_proto.envs[k] = str(v)
        return info_proto

    async def stream(
        self, request_iterator, context=None, *args, **kwargs
    ) -> AsyncIterator['Request']:
        """
        stream requests from client iterator and stream responses back.

        :param request_iterator: iterator of requests
        :param context: context of the grpc call
        :param args: positional arguments
        :param kwargs: keyword arguments
        :yield: responses to the request after streaming to Executors in Flow
        """
        self.logger.debug('recv a stream request')
        async for resp in self.streamer.rpc_stream(
            request_iterator=request_iterator, context=context, *args, **kwargs
        ):
            yield resp

    async def stream_doc(
        self, request: SingleDocumentRequest, context: 'grpc.aio.ServicerContext'
    ) -> SingleDocumentRequest:
        """
        Process the received requests and return the result as a new request

        :param request: the data request to process
        :param context: grpc context
        :yields: the response request
        """
        self.logger.debug('recv a stream_doc request')
        async for result in self.streamer.rpc_stream_doc(
            request=request,
        ):
            yield result

    async def process_single_data(
        self, request: DataRequest, context=None
    ) -> DataRequest:
        """Implements request and response handling of a single DataRequest
        :param request: DataRequest from Client
        :param context: grpc context
        :return: response DataRequest
        """
        self.logger.debug(f'recv a process_single_data request')
        return await self.streamer.process_single_data(request, context)

    async def endpoint_discovery(self, empty, context) -> jina_pb2.EndpointsProto:
        """
        Uses the connection pool to send a discover endpoint call to the Executors

        :param empty: The service expects an empty protobuf message
        :param context: grpc context
        :returns: the response request
        """
        from google.protobuf import json_format

        self.logger.debug('got an endpoint discovery request')
        response = jina_pb2.EndpointsProto()
        await self.streamer._get_endpoints_input_output_models(is_cancel=None)
        request_models_map = self.streamer._endpoints_models_map
        if request_models_map is not None and len(request_models_map) > 0:
            schema_maps = {}
            for k, v in request_models_map.items():
                schema_maps[k] = {}
                schema_maps[k]['input'] = v['input'].schema()
                schema_maps[k]['output'] = v['output'].schema()
                schema_maps[k]['is_generator'] = v['is_generator']
                schema_maps[k]['is_singleton_doc'] = v['is_singleton_doc']
            response.endpoints.extend(schema_maps.keys())
            json_format.ParseDict(schema_maps, response.schemas)
        else:
            endpoints = await self.streamer.topology_graph._get_all_endpoints(
                self.streamer._connection_pool, retry_forever=True, is_cancel=None
            )
            response.endpoints.extend(list(endpoints))
        return response

    Call = stream



"""Argparser module for container runtimes"""

import argparse

from jina.enums import DockerNetworkMode
from jina.parsers.helper import _SHOW_ALL_ARGS, KVAppendAction, add_arg_group


def mixin_container_runtime_parser(parser, pod_type: str = 'executor'):
    """Mixing in arguments required by :class:`ContainerRuntime` into the given parser.
    :param parser: the parser instance to which we add arguments
    :param pod_type: the pod_type configured by the parser. Can be either 'executor' for an Executor pod or 'gateway' for a Gateway pod
    """
    gp = add_arg_group(parser, title='ContainerRuntime')

    gp.add_argument(
        '--entrypoint',
        type=str,
        help='The entrypoint command overrides the ENTRYPOINT in Docker image. '
        'when not set then the Docker image ENTRYPOINT takes effective.',
    )
    gp.add_argument(
        '--docker-kwargs',
        action=KVAppendAction,
        metavar='KEY: VALUE',
        nargs='*',
        help='''
Dictionary of kwargs arguments that will be passed to Docker SDK when starting the docker '
container. 

More details can be found in the Docker SDK docs:  https://docker-py.readthedocs.io/en/stable/

''',
    )

    if pod_type == 'executor':
        gp.add_argument(
            '--volumes',
            type=str,
            nargs='*',
            metavar='DIR',
            help='''
The path on the host to be mounted inside the container. 

Note, 
- If separated by `:`, then the first part will be considered as the local host path and the second part is the path in the container system. 
- If no split provided, then the basename of that directory will be mounted into container's root path, e.g. `--volumes="/user/test/my-workspace"` will be mounted into `/my-workspace` inside the container. 
- All volumes are mounted with read-write mode.
    ''',
        )
        gp.add_argument(
            '--gpus',
            type=str,
            help=f'''
    This argument allows dockerized Jina Executors to discover local gpu devices.
    
    Note, 
    - To access all gpus, use `--gpus all`.
    - To access multiple gpus, e.g. make use of 2 gpus, use `--gpus 2`.
    - To access specified gpus based on device id, use `--gpus device=[YOUR-GPU-DEVICE-ID]`
    - To access specified gpus based on multiple device id, use `--gpus device=[YOUR-GPU-DEVICE-ID1],device=[YOUR-GPU-DEVICE-ID2]`
    - To specify more parameters, use `--gpus device=[YOUR-GPU-DEVICE-ID],runtime=nvidia,capabilities=display
        ''',
        )

        gp.add_argument(
            '--disable-auto-volume',
            action='store_true',
            default=False,
            help=f'Do not automatically mount a volume for dockerized Executors.',
        )

        gp.add_argument(
            '--force-network-mode',
            type=DockerNetworkMode.from_string,
            choices=list(DockerNetworkMode),
            default=DockerNetworkMode.AUTO,
            help=(
                f'''
    Force the use of the specified docker network mode (default: auto). 

    Valid options are,
    - {str(DockerNetworkMode.AUTO)}: Automatically detect the docker network.
    - {str(DockerNetworkMode.HOST)}: Use the host network.
    - {str(DockerNetworkMode.BRIDGE)}: Use a user-defined bridge network.
    - {str(DockerNetworkMode.NONE)}: Use no network (equivalent to the --network=none option).
        '''
                if _SHOW_ALL_ARGS
                else argparse.SUPPRESS
            ),
        )



from jina.jaml import JAMLCompatible
from jina.serve.helper import store_init_kwargs, wrap_func
from jina.serve.runtimes.servers import BaseServer

__all__ = ['BaseGateway']


class GatewayType(type(JAMLCompatible), type):
    """The class of Gateway type, which is the metaclass of :class:`BaseGateway`."""

    def __new__(cls, *args, **kwargs):
        """
        # noqa: DAR101
        # noqa: DAR102

        :return: Gateway class
        """
        _cls = super().__new__(cls, *args, **kwargs)
        return cls.register_class(_cls)

    @staticmethod
    def register_class(cls):
        """
        Register a class.

        :param cls: The class.
        :return: The class, after being registered.
        """
        reg_cls_set = getattr(cls, '_registered_class', set())

        cls_id = f'{cls.__module__}.{cls.__name__}'
        if cls_id not in reg_cls_set:
            reg_cls_set.add(cls_id)
            setattr(cls, '_registered_class', reg_cls_set)
            wrap_func(
                cls,
                ['__init__'],
                store_init_kwargs,
                taboo={'self', 'args', 'kwargs', 'runtime_args'},
            )
        return cls


class BaseGateway(JAMLCompatible, metaclass=GatewayType):
    """
    The base class of all custom Gateways, can be used to build a custom interface to a Jina Flow that supports
    gateway logic
    """


class Gateway(BaseServer, BaseGateway):
    """
    The class for where to inherit when you want to customize your Gateway. Important to provide backwards compatibility
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



from jina import Executor, requests
from docarray import DocList
from docarray.documents import TextDoc


class MyExecutor(Executor):
    @requests
    def foo(self, docs: DocList[TextDoc], **kwargs) -> DocList[TextDoc]:
        docs[0].text = 'hello, world!'
        docs[1].text = 'goodbye, world!'
        return docs



import asyncio
from typing import TYPE_CHECKING, Optional, Tuple

import grpc

from jina.clients.base.retry import wait_or_raise_err
from jina.clients.helper import callback_exec
from jina.excepts import InternalNetworkError
from jina.proto import jina_pb2_grpc
from jina.serve.stream import RequestStreamer

if TYPE_CHECKING:
    from jina.types.request import Request


class UnaryRpc:
    """Class that encapsulated the methods required to run unary rpc calls from the client. Instantiate a single class
    for each client request.
    """

    def __init__(
        self,
        channel,
        continue_on_error,
        metadata,
        on_always,
        on_done,
        on_error,
        p_bar,
        req_iter,
        max_attempts,
        backoff_multiplier,
        initial_backoff,
        max_backoff,
        logger,
        show_progress,
        compression,
        client_args,
        prefetch,
        results_in_order,
        **kwargs
    ):
        self.results_in_order = results_in_order
        self.prefetch = prefetch
        self.client_args = client_args
        self.compression = compression
        self.show_progress = show_progress
        self.logger = logger
        self.max_backoff = max_backoff
        self.initial_backoff = initial_backoff
        self.backoff_multiplier = backoff_multiplier
        self.max_attempts = max_attempts
        self.req_iter = req_iter
        self.p_bar = p_bar
        self.on_error = on_error
        self.on_done = on_done
        self.on_always = on_always
        self.metadata = metadata
        self.continue_on_error = continue_on_error
        self.channel = channel
        self.kwargs = kwargs

    async def unary_rpc_with_retry(self):
        """Wraps the unary rpc call with retry loop based on the retry params.
        :yields: Responses received from the target.
        """
        stub = jina_pb2_grpc.JinaSingleDataRequestRPCStub(self.channel)

        def _request_handler(
            request: 'Request', **kwargs
        ) -> 'Tuple[asyncio.Future, Optional[asyncio.Future]]':
            async def _with_retry(req: 'Request'):
                for attempt in range(1, self.max_attempts + 1):
                    try:
                        return await stub.process_single_data(
                            req,
                            compression=self.compression,
                            metadata=self.metadata,
                            credentials=self.kwargs.get('credentials', None),
                            timeout=self.kwargs.get('timeout', None),
                        )
                    except (
                        grpc.aio.AioRpcError,
                        InternalNetworkError,
                    ) as err:
                        await wait_or_raise_err(
                            attempt=attempt,
                            err=err,
                            max_attempts=self.max_attempts,
                            backoff_multiplier=self.backoff_multiplier,
                            initial_backoff=self.initial_backoff,
                            max_backoff=self.max_backoff,
                        )

            return (
                asyncio.ensure_future(_with_retry(request)),
                None,
            )

        def _result_handler(resp):
            callback_exec(
                response=resp,
                logger=self.logger,
                on_error=self.on_error,
                on_done=self.on_done,
                on_always=self.on_always,
                continue_on_error=self.continue_on_error,
            )
            return resp

        streamer_args = vars(self.client_args)
        if self.prefetch:
            streamer_args['prefetch'] = self.prefetch
        streamer = RequestStreamer(
            request_handler=_request_handler,
            result_handler=_result_handler,
            iterate_sync_in_thread=False,
            logger=self.logger,
            **streamer_args,
        )
        async for response in streamer.stream(
            request_iterator=self.req_iter, results_in_order=self.results_in_order
        ):
            if self.show_progress:
                self.p_bar.update()
            yield response



import random
import asyncio
import time

import aiohttp
import grpc


def _raise_last_attempt(err, attempt):
    if isinstance(err, asyncio.CancelledError):
        trailing_metadata = grpc.aio.Metadata()
        trailing_metadata.add('jina-client-attempts', str(attempt))
        raise grpc.aio.AioRpcError(
            code=grpc.StatusCode.CANCELLED,
            initial_metadata=grpc.aio.Metadata(),
            trailing_metadata=trailing_metadata,
        )
    elif isinstance(err, grpc.aio.AioRpcError):
        trailing_metadata = err.trailing_metadata() or grpc.aio.Metadata()
        trailing_metadata.add('jina-client-attempts', str(attempt))
        raise grpc.aio.AioRpcError(
            code=err.code(),
            details=err.details(),
            initial_metadata=err.initial_metadata(),
            trailing_metadata=trailing_metadata,
            debug_error_string=err.debug_error_string(),
        )
    elif isinstance(err, aiohttp.ClientConnectorCertificateError):
        raise err
    elif isinstance(err, aiohttp.ClientError):
        raise ConnectionError(str(err))
    else:
        raise err


def sync_wait_or_raise_err(
    attempt: int,
    err: Exception,
    max_attempts: float,
    backoff_multiplier: float,
    initial_backoff: float,
    max_backoff: float,
):
    """
    Accepts retry parameters and the underlying. The error is raised if the max_attempts has been reached otherwise the
    method waits based on the backoff calculations.
    :param attempt: Number of the current attempt.
    :param err: Underlying error that was raised by the operation.
    :param max_attempts: Maximum number of attempts that are allowed.
    :param backoff_multiplier: Factor that will be raised to the exponent of (attempt - 1) for calculating the backoff wait time.
    :param initial_backoff: The backoff time on the first error. This will be multiplied by the backoff_multiplier exponent for subsequent wait time calculations.
    :param max_backoff: The maximum backoff wait time.
    """
    if attempt == max_attempts:
        _raise_last_attempt(err, attempt)
    else:
        time.sleep(
            _wait_time(attempt, backoff_multiplier, initial_backoff, max_backoff)
        )


async def wait_or_raise_err(
    attempt: int,
    err: Exception,
    max_attempts: float,
    backoff_multiplier: float,
    initial_backoff: float,
    max_backoff: float,
):
    """
    Accepts retry parameters and the underlying. The error is raised if the max_attempts has been reached otherwise the
    method waits based on the backoff calculations.
    :param attempt: Number of the current attempt.
    :param err: Underlying error that was raised by the operation.
    :param max_attempts: Maximum number of attempts that are allowed.
    :param backoff_multiplier: Factor that will be raised to the exponent of (attempt - 1) for calculating the backoff wait time.
    :param initial_backoff: The backoff time on the first error. This will be multiplied by the backoff_multiplier exponent for subsequent wait time calculations.
    :param max_backoff: The maximum backoff wait time.
    """

    if attempt == max_attempts:
        _raise_last_attempt(err, attempt)
    else:
        await asyncio.sleep(
            _wait_time(attempt, backoff_multiplier, initial_backoff, max_backoff)
        )


def _wait_time(attempt, backoff_multiplier, initial_backoff, max_backoff):
    if attempt == 1:
        wait_time = initial_backoff
    else:
        wait_time = random.uniform(
            0,
            min(initial_backoff * backoff_multiplier ** (attempt - 1), max_backoff),
        )
    return wait_time



from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.servers.load_balancer import LoadBalancingServer

__all__ = ['LoadBalancerGateway']


class LoadBalancerGateway(LoadBalancingServer, BaseGateway):
    """
    :class:`LoadBalancerGateway`
    """

    pass



"""Helper functions for clients in Jina."""

from functools import wraps
from typing import Callable, Optional

from jina.excepts import BadClientCallback, BadServer
from jina.helper import get_rich_console
from jina.logging.logger import JinaLogger
from jina.proto import jina_pb2
from jina.types.request.data import Response


def pprint_routes(resp: 'Response', stack_limit: int = 3):
    """Pretty print routes with :mod:`prettytable`, fallback to :func:`print`.

    :param resp: the :class:`Response` object
    :param stack_limit: traceback limit
    """
    routes = resp.routes

    from rich import box
    from rich.table import Table

    table = Table(box=box.SIMPLE)
    for v in ('Executor', 'Time', 'Exception'):
        table.add_column(v)

    for route in routes:
        status_icon = '🟢'
        if route.status.code == jina_pb2.StatusProto.ERROR:
            status_icon = '🔴'

        table.add_row(
            f'{status_icon} {route.executor}',
            f'{route.start_time.ToMilliseconds() - routes[0].start_time.ToMilliseconds()}ms',
            ''.join(route.status.exception.stacks[-stack_limit:]),
        )

    console = get_rich_console()
    console.print(table)


def _safe_callback(func: Callable, continue_on_error: bool, logger) -> Callable:
    @wraps(func)
    def _arg_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            err_msg = f'uncaught exception in callback {func.__name__}(): {ex!r}'
            if continue_on_error:
                logger.error(err_msg)
            else:
                raise BadClientCallback(err_msg) from ex

    return _arg_wrapper


def callback_exec(
    response,
    logger: JinaLogger,
    on_done: Optional[Callable] = None,
    on_error: Optional[Callable] = None,
    on_always: Optional[Callable] = None,
    continue_on_error: bool = False,
) -> None:
    """Execute the callback with the response.

    :param response: the response
    :param on_done: the on_done callback
    :param on_error: the on_error callback
    :param on_always: the on_always callback
    :param continue_on_error: whether to continue on error
    :param logger: a logger instance
    """
    if response.header.status.code >= jina_pb2.StatusProto.ERROR:
        if on_error:
            _safe_callback(on_error, continue_on_error, logger)(response)
        elif continue_on_error:
            logger.error(f'Server error: {response.header}')
        else:
            raise BadServer(response.header)
    elif on_done and response.header.status.code == jina_pb2.StatusProto.SUCCESS:
        _safe_callback(on_done, continue_on_error, logger)(response)
    if on_always:
        _safe_callback(on_always, continue_on_error, logger)(response)



"""Argparser module for Pod runtimes"""

import argparse
from dataclasses import dataclass
from typing import Dict

from jina.enums import PodRoleType, ProtocolType, ProviderType
from jina.helper import random_port
from jina.parsers.helper import (
    _SHOW_ALL_ARGS,
    CastPeerPorts,
    CastToIntAction,
    KVAppendAction,
    add_arg_group,
)


@dataclass
class PodTypeParams:
    """Data Class representing possible parameters for each pod type"""

    runtime_cls: str
    role_type: PodRoleType


POD_PARAMS_MAPPING: Dict[str, PodTypeParams] = {
    'worker': PodTypeParams(runtime_cls='WorkerRuntime', role_type=PodRoleType.WORKER),
    'head': PodTypeParams(runtime_cls='HeadRuntime', role_type=PodRoleType.HEAD),
    'gateway': PodTypeParams(
        runtime_cls='GatewayRuntime', role_type=PodRoleType.GATEWAY
    ),
}


def mixin_pod_parser(parser, pod_type: str = 'worker'):
    """Mixing in arguments required by :class:`Pod` into the given parser.
    :param parser: the parser instance to which we add arguments
    :param pod_type: the pod_type configured by the parser. Can be either 'worker' for WorkerRuntime or 'gateway' for GatewayRuntime
    """

    gp = add_arg_group(parser, title='Pod')

    gp.add_argument(
        '--runtime-cls',
        type=str,
        default=POD_PARAMS_MAPPING[pod_type].runtime_cls,
        help='The runtime class to run inside the Pod',
    )

    gp.add_argument(
        '--timeout-ready',
        type=int,
        default=600000,
        help='The timeout in milliseconds of a Pod waits for the runtime to be ready, -1 for waiting '
        'forever',
    )

    gp.add_argument(
        '--env',
        action=KVAppendAction,
        metavar='KEY: VALUE',
        nargs='*',
        help='The map of environment variables that are available inside runtime',
    )

    gp.add_argument(
        '--env-from-secret',
        action=KVAppendAction,
        metavar='KEY: VALUE',
        nargs='*',
        help=(
            'The map of environment variables that are read from kubernetes cluster secrets'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )
    gp.add_argument(
        '--image-pull-secrets',
        type=str,
        nargs='+',
        default=None,
        help=(
            'List of ImagePullSecrets that the Kubernetes Pods need to have access to in order to pull the image. Used in `to_kubernetes_yaml`'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )

    # hidden CLI used for internal only

    gp.add_argument(
        '--shard-id',
        type=int,
        default=0,
        help=(
            'defines the shard identifier for the executor. It is used as suffix for the workspace path of the executor`'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )

    gp.add_argument(
        '--pod-role',
        type=PodRoleType.from_string,
        choices=list(PodRoleType),
        default=POD_PARAMS_MAPPING[pod_type].role_type,
        help=(
            'The role of this Pod in a Deployment'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )

    gp.add_argument(
        '--noblock-on-start',
        action='store_true',
        default=False,
        help=(
            'If set, starting a Pod/Deployment does not block the thread/process. It then relies on '
            '`wait_start_success` at outer function for the postpone check.'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )

    gp.add_argument(
        '--floating',
        action='store_true',
        default=False,
        help='If set, the current Pod/Deployment can not be further chained, '
        'and the next `.add()` will chain after the last Pod/Deployment not this current one.',
    )

    gp.add_argument(
        '--replica-id',
        type=int,
        default=0,
        help=(
            'defines the replica identifier for the executor. It is used when `stateful` is set to true'
            if _SHOW_ALL_ARGS
            else argparse.SUPPRESS
        ),
    )

    if pod_type != 'gateway':
        gp.add_argument(
            '--reload',
            action='store_true',
            default=False,
            help='If set, the Executor will restart while serving if YAML configuration source or Executor modules '
            'are changed. If YAML configuration is changed, the whole deployment is reloaded and new '
            'processes will be restarted. If only Python modules of the Executor have changed, they will be '
            'reloaded to the interpreter without restarting process.',
        )
        gp.add_argument(
            '--install-requirements',
            action='store_true',
            default=False,
            help='If set, try to install `requirements.txt` from the local Executor if exists in the Executor folder. If using Hub, install `requirements.txt` in the Hub Executor bundle to local.',
        )
    else:
        gp.add_argument(
            '--reload',
            action='store_true',
            default=False,
            help='If set, the Gateway will restart while serving if YAML configuration source is changed.',
        )
    mixin_pod_runtime_args_parser(gp, pod_type=pod_type)
    mixin_stateful_parser(gp)


def mixin_pod_runtime_args_parser(arg_group, pod_type='worker'):
    """Mixin for runtime arguments of pods
    :param arg_group: the parser instance or args group to which we add arguments
    :param pod_type: the pod_type configured by the parser. Can be either 'worker' for WorkerRuntime or 'gateway' for GatewayRuntime
    """
    alias = ['--port', '--ports']
    if pod_type != 'gateway':
        port_description = (
            'The port for input data to bind to, default is a random port between [49152, 65535]. '
            'In the case of an external Executor (`--external` or `external=True`) this can be a list of ports. '
            'Then, every resulting address will be considered as one replica of the Executor.'
        )
    else:
        port_description = (
            'The port for input data to bind the gateway server to, by default, random ports between range [49152, 65535] will be assigned. '
            'The port argument can be either 1 single value in case only 1 protocol is used or multiple values when '
            'many protocols are used.'
        )
        alias.extend(['--port-expose', '--port-in'])

    arg_group.add_argument(
        *alias,
        action=CastToIntAction,
        type=str,
        nargs='+',
        default=[random_port()],
        help=port_description,
    )

    server_name = 'Gateway' if pod_type == 'gateway' else 'Executor'
    arg_group.add_argument(
        '--protocol',
        '--protocols',
        nargs='+',
        type=ProtocolType.from_string,
        choices=list(ProtocolType),
        default=[ProtocolType.GRPC],
        help=f'Communication protocol of the server exposed by the {server_name}. This can be a single value or a list of protocols, depending on your chosen Gateway. Choose the convenient protocols from: {[protocol.to_string() for protocol in list(ProtocolType)]}.',
    )

    arg_group.add_argument(
        '--provider',
        type=ProviderType.from_string,
        choices=list(ProviderType),
        default=[ProviderType.NONE],
        help=f'If set, Executor is translated to a custom container compatible with the chosen provider. Choose the convenient providers from: {[provider.to_string() for provider in list(ProviderType)]}.',
    )

    arg_group.add_argument(
        '--provider-endpoint',
        type=str,
        default=None,
        help=f'If set, Executor endpoint will be explicitly chosen and used in the custom container operated by the provider.',
    )

    arg_group.add_argument(
        '--monitoring',
        action='store_true',
        default=False,
        help='If set, spawn an http server with a prometheus endpoint to expose metrics',
    )

    arg_group.add_argument(
        '--port-monitoring',
        type=str,
        nargs='+',
        default=[random_port()],
        action=CastToIntAction,
        dest='port_monitoring',
        help=f'The port on which the prometheus server is exposed, default is a random port between [49152, 65535]',
    )

    arg_group.add_argument(
        '--retries',
        type=int,
        default=-1,
        dest='retries',
        help=f'Number of retries per gRPC call. If <0 it defaults to max(3, num_replicas)',
    )

    arg_group.add_argument(
        '--tracing',
        action='store_true',
        default=False,
        help='If set, the sdk implementation of the OpenTelemetry tracer will be available and will be enabled for automatic tracing of requests and customer span creation. '
        'Otherwise a no-op implementation will be provided.',
    )

    arg_group.add_argument(
        '--traces-exporter-host',
        type=str,
        default=None,
        help='If tracing is enabled, this hostname will be used to configure the trace exporter agent.',
    )

    arg_group.add_argument(
        '--traces-exporter-port',
        type=int,
        default=None,
        help='If tracing is enabled, this port will be used to configure the trace exporter agent.',
    )

    arg_group.add_argument(
        '--metrics',
        action='store_true',
        default=False,
        help='If set, the sdk implementation of the OpenTelemetry metrics will be available for default monitoring and custom measurements. '
        'Otherwise a no-op implementation will be provided.',
    )

    arg_group.add_argument(
        '--metrics-exporter-host',
        type=str,
        default=None,
        help='If tracing is enabled, this hostname will be used to configure the metrics exporter agent.',
    )

    arg_group.add_argument(
        '--metrics-exporter-port',
        type=int,
        default=None,
        help='If tracing is enabled, this port will be used to configure the metrics exporter agent.',
    )


def mixin_stateful_parser(parser):
    """Mixing in arguments required to work with Stateful Executors into the given parser.
    :param parser: the parser instance to which we add arguments
    """

    gp = add_arg_group(parser, title='Stateful Executor')

    gp.add_argument(
        '--stateful',
        action='store_true',
        default=False,
        help='If set, start consensus module to make sure write operations are properly replicated between all the replicas',
    )
    gp.add_argument(
        '--peer-ports',
        type=str,
        default=None,
        help='When using --stateful option, it is required to tell the cluster what are the cluster configuration. This is important'
        'when the Deployment is restarted. It indicates the ports to which each replica of the cluster binds.'
        ' It is expected to be a single list if shards == 1 or a dictionary if shards > 1.',
        action=CastPeerPorts,
        nargs='+',
    )



import copy
import os
from argparse import Namespace
from typing import Dict, List, Optional, Tuple, Union

from jina.constants import (
    __default_composite_gateway__,
    __default_executor__,
    __default_grpc_gateway__,
    __default_http_gateway__,
    __default_websocket_gateway__,
)
from jina.enums import PodRoleType
from jina.excepts import NoContainerizedError
from jina.orchestrate.deployments import Deployment
from jina.orchestrate.deployments.config.helper import (
    construct_runtime_container_args,
    get_base_executor_version,
    get_image_name,
    to_compatible_name,
    validate_uses,
)
from jina.orchestrate.helper import generate_default_volume_and_workspace
from jina.orchestrate.deployments.config.helper import resolve_image_name

port = 8081


class DockerComposeConfig:
    """
    Class that implements the output of configuration files for docker-compose for a given Deployment.
    """

    class _DockerComposeService:
        def __init__(
            self,
            name: str,
            version: str,
            pod_type: PodRoleType,
            jina_deployment_name: str,
            shard_id: Optional[int],
            common_args: Union['Namespace', Dict],
            service_args: Union['Namespace', Dict],
            deployments_addresses: Optional[Dict[str, List[str]]] = None,
        ):
            self.name = name
            self.compatible_name = to_compatible_name(self.name)
            self.version = version
            self.pod_type = pod_type
            self.jina_deployment_name = jina_deployment_name
            self.shard_id = shard_id
            self.common_args = common_args
            self.service_args = service_args
            self.num_replicas = getattr(self.service_args, 'replicas', 1)
            self.deployments_addresses = deployments_addresses

        def get_gateway_config(
            self,
        ) -> Dict:
            import os

            cargs = copy.copy(self.service_args)

            image_name = resolve_image_name(cargs.uses)

            cargs.deployments_addresses = self.deployments_addresses
            from jina.helper import ArgNamespace
            from jina.parsers import set_gateway_parser

            taboo = {
                'uses_metas',
                'volumes',
                'uses_before',
                'uses_after',
                'workspace',
                'workspace_id',
                'noblock_on_start',
                'env',
            }
            if cargs.uses not in [
                __default_http_gateway__,
                __default_websocket_gateway__,
                __default_grpc_gateway__,
                __default_composite_gateway__,
            ]:
                cargs.uses = 'config.yml'

            non_defaults = ArgNamespace.get_non_defaults_args(
                cargs, set_gateway_parser(), taboo=taboo
            )
            _args = ArgNamespace.kwargs2list(non_defaults)

            container_args = ['gateway'] + _args

            protocol = str(non_defaults.get('protocol', ['grpc'])[0]).lower()

            ports = cargs.port + ([cargs.port_monitoring] if cargs.monitoring else [])

            envs = [f'JINA_LOG_LEVEL={os.getenv("JINA_LOG_LEVEL", "INFO")}']
            if cargs.env:
                for k, v in cargs.env.items():
                    envs.append(f'{k}={v}')
            return {
                'image': image_name,
                'entrypoint': ['jina'],
                'command': container_args,
                'expose': ports,
                'ports': [f'{_port}:{_port}' for _port in ports],
                'healthcheck': {
                    'test': f'jina ping gateway {protocol}://127.0.0.1:{cargs.port[0]}',
                    'interval': '2s',
                },
                'environment': envs,
            }

        def _get_container_args(self, cargs):
            uses_metas = cargs.uses_metas or {}
            uses_with = self.service_args.uses_with
            if cargs.uses != __default_executor__:
                cargs.uses = 'config.yml'
            return construct_runtime_container_args(
                cargs, uses_metas, uses_with, self.pod_type
            )

        def _update_config_with_volumes(self, config, auto_volume=True):
            if self.service_args.volumes:  # respect custom volume definition
                config['volumes'] = self.service_args.volumes
                return config

            if not auto_volume:
                return config

            # if no volume is given, create default volume
            (
                generated_volumes,
                workspace_in_container,
            ) = generate_default_volume_and_workspace(
                workspace_id=self.service_args.workspace_id
            )
            config['volumes'] = generated_volumes
            if (
                '--workspace' not in config['command']
            ):  # set workspace only of not already given
                config['command'].append('--workspace')
                config['command'].append(workspace_in_container)
            return config

        def get_runtime_config(self) -> List[Dict]:
            # One Dict for replica
            replica_configs = []
            for i_rep in range(self.service_args.replicas):
                cargs = copy.copy(self.service_args)
                cargs.name = (
                    f'{cargs.name}/rep-{i_rep}'
                    if self.service_args.replicas > 1
                    else cargs.name
                )

                env = cargs.env
                image_name = resolve_image_name(cargs.uses)
                container_args = self._get_container_args(cargs)
                config = {
                    'image': image_name,
                    'entrypoint': ['jina'],
                    'command': container_args,
                    'healthcheck': {
                        'test': f'jina ping executor 127.0.0.1:{cargs.port}',
                        'interval': '2s',
                    },
                    'environment': [
                        f'JINA_LOG_LEVEL={os.getenv("JINA_LOG_LEVEL", "INFO")}'
                    ],
                }

                if cargs.gpus:
                    try:
                        count = int(cargs.gpus)
                    except ValueError:
                        count = cargs.gpus

                    config['deploy'] = {
                        'resources': {
                            'reservations': {
                                'devices': [
                                    {
                                        'driver': 'nvidia',
                                        'count': count,
                                        'capabilities': ['gpu'],
                                    }
                                ]
                            }
                        }
                    }

                if cargs.monitoring:
                    config['expose'] = [cargs.port_monitoring]
                    config['ports'] = [
                        f'{cargs.port_monitoring}:{cargs.port_monitoring}'
                    ]

                if env is not None:
                    config['environment'] = [f'{k}={v}' for k, v in env.items()]

                if self.service_args.pod_role == PodRoleType.WORKER:
                    config = self._update_config_with_volumes(
                        config, auto_volume=not self.common_args.disable_auto_volume
                    )

                replica_configs.append(config)
            return replica_configs

    def __init__(
        self,
        args: Union['Namespace', Dict],
        deployments_addresses: Optional[Dict[str, List[str]]] = None,
    ):
        if not validate_uses(args.uses):
            raise NoContainerizedError(
                f'Executor "{args.uses}" is not valid to be used in docker-compose. '
                'You need to use a containerized Executor. You may check `jina hub --help` to see how Jina Hub can help you building containerized Executors.'
            )
        self.deployments_addresses = deployments_addresses
        self.head_service = None
        self.uses_before_service = None
        self.uses_after_service = None
        self.args = copy.copy(args)
        self.name = self.args.name

        self.services_args = self._get_services_args(self.args)

        if self.services_args['head_service'] is not None:
            self.head_service = self._DockerComposeService(
                name=self.services_args['head_service'].name,
                version=get_base_executor_version(),
                shard_id=None,
                jina_deployment_name=self.name,
                common_args=self.args,
                service_args=self.services_args['head_service'],
                pod_type=PodRoleType.HEAD,
                deployments_addresses=None,
            )

        if self.services_args['uses_before_service'] is not None:
            self.uses_before_service = self._DockerComposeService(
                name=self.services_args['uses_before_service'].name,
                version=get_base_executor_version(),
                shard_id=None,
                jina_deployment_name=self.name,
                common_args=self.args,
                service_args=self.services_args['uses_before_service'],
                pod_type=PodRoleType.WORKER,
                deployments_addresses=None,
            )

        if self.services_args['uses_after_service'] is not None:
            self.uses_after_service = self._DockerComposeService(
                name=self.services_args['uses_after_service'].name,
                version=get_base_executor_version(),
                shard_id=None,
                jina_deployment_name=self.name,
                common_args=self.args,
                service_args=self.services_args['uses_after_service'],
                pod_type=PodRoleType.WORKER,
                deployments_addresses=None,
            )

        self.worker_services = []
        services_args = self.services_args['services']
        for i, args in enumerate(services_args):
            name = f'{self.name}-{i}' if len(services_args) > 1 else f'{self.name}'
            self.worker_services.append(
                self._DockerComposeService(
                    name=name,
                    version=get_base_executor_version(),
                    shard_id=i,
                    common_args=self.args,
                    service_args=args,
                    pod_type=(
                        PodRoleType.WORKER if name != 'gateway' else PodRoleType.GATEWAY
                    ),
                    jina_deployment_name=self.name,
                    deployments_addresses=(
                        self.deployments_addresses if name == 'gateway' else None
                    ),
                )
            )

    def _get_services_args(self, args):
        parsed_args = {
            'head_service': None,
            'uses_before_service': None,
            'uses_after_service': None,
            'services': [],
        }
        shards = getattr(args, 'shards', 1)
        replicas = getattr(args, 'replicas', 1)
        uses_before = getattr(args, 'uses_before', None)
        uses_after = getattr(args, 'uses_after', None)

        if args.name != 'gateway' and shards > 1:
            parsed_args['head_service'] = Deployment._copy_to_head_args(self.args)
            parsed_args['head_service'].port = port
            parsed_args['head_service'].uses = None
            parsed_args['head_service'].uses_metas = None
            parsed_args['head_service'].uses_with = None
            parsed_args['head_service'].uses_before = None
            parsed_args['head_service'].uses_after = None

            # if the k8s connection pool is disabled, the connection pool is managed manually
            import json

            connection_list = {}
            for shard_id in range(shards):
                shard_name = f'{self.name}-{shard_id}' if shards > 1 else f'{self.name}'
                connection_list[str(shard_id)] = []
                for i_rep in range(replicas):
                    replica_name = (
                        f'{shard_name}/rep-{i_rep}' if replicas > 1 else shard_name
                    )
                    connection_list[str(shard_id)].append(
                        f'{to_compatible_name(replica_name)}:{port}'
                    )

            parsed_args['head_service'].connection_list = json.dumps(connection_list)

        if uses_before and shards > 1:
            uses_before_cargs = copy.deepcopy(args)
            uses_before_cargs.shard_id = 0
            uses_before_cargs.replicas = 1
            uses_before_cargs.name = f'{args.name}/uses-before'
            uses_before_cargs.uses = args.uses_before
            uses_before_cargs.uses_before = None
            uses_before_cargs.uses_after = None
            uses_before_cargs.uses_with = None
            uses_before_cargs.uses_metas = None
            uses_before_cargs.env = None
            uses_before_cargs.host = args.host[0]
            uses_before_cargs.port = port
            uses_before_cargs.uses_before_address = None
            uses_before_cargs.uses_after_address = None
            uses_before_cargs.connection_list = None
            uses_before_cargs.pod_role = PodRoleType.WORKER
            uses_before_cargs.polling = None
            parsed_args['uses_before_service'] = uses_before_cargs
            parsed_args['head_service'].uses_before_address = (
                f'{to_compatible_name(uses_before_cargs.name)}:{uses_before_cargs.port}'
            )
        if uses_after and shards > 1:
            uses_after_cargs = copy.deepcopy(args)
            uses_after_cargs.shard_id = 0
            uses_after_cargs.replicas = 1
            uses_after_cargs.name = f'{args.name}/uses-after'
            uses_after_cargs.uses = args.uses_after
            uses_after_cargs.uses_before = None
            uses_after_cargs.uses_after = None
            uses_after_cargs.uses_with = None
            uses_after_cargs.uses_metas = None
            uses_after_cargs.env = None
            uses_after_cargs.host = args.host[0]
            uses_after_cargs.port = port
            uses_after_cargs.uses_before_address = None
            uses_after_cargs.uses_after_address = None
            uses_after_cargs.connection_list = None
            uses_after_cargs.pod_role = PodRoleType.WORKER
            uses_after_cargs.polling = None
            parsed_args['uses_after_service'] = uses_after_cargs
            parsed_args['head_service'].uses_after_address = (
                f'{to_compatible_name(uses_after_cargs.name)}:{uses_after_cargs.port}'
            )

        for i in range(shards):
            cargs = copy.deepcopy(args)
            cargs.shard_id = i
            cargs.uses_before = None
            cargs.uses_after = None
            cargs.uses_before_address = None
            cargs.uses_after_address = None
            if shards > 1:
                cargs.name = f'{cargs.name}-{i}'
            if args.name == 'gateway':
                cargs.pod_role = PodRoleType.GATEWAY
            else:
                cargs.port = port
                cargs.host = args.host[0]
            parsed_args['services'].append(cargs)

        return parsed_args

    def to_docker_compose_config(
        self,
    ) -> List[Tuple[str, Dict]]:
        """
        Return a list of dictionary configurations. One for each service in this Deployment
            .. # noqa: DAR201
            .. # noqa: DAR101
        """
        if self.name == 'gateway':
            return [
                (
                    'gateway',
                    self.worker_services[0].get_gateway_config(),
                )
            ]
        else:
            services = []
            if self.head_service is not None:
                services.append(
                    (
                        self.head_service.compatible_name,
                        self.head_service.get_runtime_config()[0],
                    )
                )
            if self.uses_before_service is not None:
                services.append(
                    (
                        self.uses_before_service.compatible_name,
                        self.uses_before_service.get_runtime_config()[0],
                    )
                )
            if self.uses_after_service is not None:
                services.append(
                    (
                        self.uses_after_service.compatible_name,
                        self.uses_after_service.get_runtime_config()[0],
                    )
                )
            for worker_service in self.worker_services:
                configs = worker_service.get_runtime_config()
                for rep_id, config in enumerate(configs):
                    name = (
                        f'{worker_service.name}/rep-{rep_id}'
                        if len(configs) > 1
                        else worker_service.name
                    )
                    services.append((to_compatible_name(name), config))
            return services



from jina.serve.runtimes.gateway.gateway import BaseGateway
from jina.serve.runtimes.servers.websocket import WebSocketServer

__all__ = ['WebSocketGateway']


class WebSocketGateway(WebSocketServer, BaseGateway):
    """
    :class:`WebSocketGateway` is a WebSocketServer that can be loaded from YAML as any other Gateway
    """

    pass



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import add_voter_pb2 as add__voter__pb2


class RaftAdminStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AddVoter = channel.unary_unary(
            '/RaftAdmin/AddVoter',
            request_serializer=add__voter__pb2.AddVoterRequest.SerializeToString,
            response_deserializer=add__voter__pb2.Future.FromString,
        )
        self.Await = channel.unary_unary(
            '/RaftAdmin/Await',
            request_serializer=add__voter__pb2.Future.SerializeToString,
            response_deserializer=add__voter__pb2.AwaitResponse.FromString,
        )
        self.Forget = channel.unary_unary(
            '/RaftAdmin/Forget',
            request_serializer=add__voter__pb2.Future.SerializeToString,
            response_deserializer=add__voter__pb2.ForgetResponse.FromString,
        )


class RaftAdminServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AddVoter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Await(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Forget(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_RaftAdminServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'AddVoter': grpc.unary_unary_rpc_method_handler(
            servicer.AddVoter,
            request_deserializer=add__voter__pb2.AddVoterRequest.FromString,
            response_serializer=add__voter__pb2.Future.SerializeToString,
        ),
        'Await': grpc.unary_unary_rpc_method_handler(
            servicer.Await,
            request_deserializer=add__voter__pb2.Future.FromString,
            response_serializer=add__voter__pb2.AwaitResponse.SerializeToString,
        ),
        'Forget': grpc.unary_unary_rpc_method_handler(
            servicer.Forget,
            request_deserializer=add__voter__pb2.Future.FromString,
            response_serializer=add__voter__pb2.ForgetResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'RaftAdmin', rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class RaftAdmin(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AddVoter(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/RaftAdmin/AddVoter',
            add__voter__pb2.AddVoterRequest.SerializeToString,
            add__voter__pb2.Future.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Await(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/RaftAdmin/Await',
            add__voter__pb2.Future.SerializeToString,
            add__voter__pb2.AwaitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Forget(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/RaftAdmin/Forget',
            add__voter__pb2.Future.SerializeToString,
            add__voter__pb2.ForgetResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""Module for helper functions for clients."""

from typing import Optional, Tuple

from jina._docarray import Document, DocumentArray, docarray_v2
from jina.enums import DataInputType
from jina.types.request.data import DataRequest

if docarray_v2:
    from docarray import DocList, BaseDoc


def _new_data_request_from_batch(
    batch,
    data_type: DataInputType,
    endpoint: str,
    target: Optional[str],
    parameters: Optional[dict],
) -> DataRequest:
    req = _new_data_request(endpoint, target, parameters)

    # add docs fields
    _add_docs(req, batch, data_type)

    return req


def _new_data_request(
    endpoint: str, target: Optional[str], parameters: Optional[dict]
) -> DataRequest:
    req = DataRequest()

    # set up header
    req.header.exec_endpoint = endpoint
    if target:
        req.header.target_executor = target
    # add parameters field
    if parameters:
        req.parameters = parameters
    return req


def _new_doc_from_data(
    data, data_type: DataInputType
) -> Tuple['Document', 'DataInputType']:
    def _build_doc_from_content():
        return Document(content=data), DataInputType.CONTENT

    if data_type == DataInputType.DICT:
        return (
            (Document(**data), DataInputType.DICT)
            if docarray_v2
            else (Document.from_dict(data), DataInputType.DICT)
        )
    if data_type == DataInputType.AUTO or data_type == DataInputType.DOCUMENT:
        if isinstance(data, Document):
            # if incoming is already primitive type Document, then all good, best practice!
            return data, DataInputType.DOCUMENT
        elif isinstance(data, dict):
            return (
                (Document(**data), DataInputType.DICT)
                if docarray_v2
                else (Document.from_dict(data), DataInputType.DICT)
            )
        else:
            try:
                d = Document(data)
                return d, DataInputType.DOCUMENT  # NOT HIT
            except ValueError:
                # AUTO has a fallback, now reconsider it as content
                if data_type == DataInputType.AUTO:
                    return _build_doc_from_content()
                else:
                    raise

    elif data_type == DataInputType.CONTENT:
        return _build_doc_from_content()


def _add_docs(req: DataRequest, batch, data_type: DataInputType) -> None:
    if not docarray_v2:
        da = DocumentArray([])
    else:
        if len(batch) > 0:
            da = DocList[batch[0].__class__]()
        else:
            da = DocList[BaseDoc]()

    for content in batch:
        d, data_type = _new_doc_from_data(content, data_type)
        da.append(d)
    req.document_array_cls = da.__class__
    req.data.docs = da
